import numpy as np
import io
from PIL import Image
import matplotlib


def get_img_from_fig(fig, dpi=180):
    buf = io.BytesIO()
    from io import BytesIO

    fig.savefig(buf, format="png", dpi=dpi, bbox_inches='tight',
                pad_inches=0, transparent=True)
    buf.seek(0)
    img_arr = np.frombuffer(buf.getvalue(), dtype=np.uint8)
    img = np.asarray(Image.open(BytesIO(img_arr)))
    buf.close()
    #img = cv2.imdecode(img_arr, 1)
    #img = cv2.cvtColor(img, cv2.COLOR_BGRA2RGB)

    return img



from anndata import AnnData
from typing import Optional, Union
from scipy.stats import pearsonr
import gseapy
import pandas as pd


def enrichr(
    adata: AnnData,
    use_data: str = "X_fa",
    factor: int = 1,
    top_genes: list = None,
    gene_sets: str = "KEGG_2019_Human",
    cutoff: float = 0.05,
    copy: bool = False
) -> Optional[AnnData]:
    
    print("Check more gene sets here: https://amp.pharm.mssm.edu/Enrichr/#stats")

    if top_genes is None:
    
        top_genes = adata.uns["factor_sig"][use_data]["Factor_" + str(factor)]["top_genes"]

        gl = gseapy.enrichr(gene_list=top_genes,
                            gene_sets=gene_sets,
                            cutoff=cutoff, outdir=None)

        result = gl.results

        result = result[result["Adjusted P-value"] < cutoff]

        adata.uns['factor_sig'][use_data]["Factor_"+str(factor)][gene_sets] = {}

        adata.uns['factor_sig'][use_data]["Factor_"+str(factor)][gene_sets].update({"result":result}) 

        print('Enrichment analysis of top genes in Factor ' + str(factor) + 
            ' is done! The result is stored in adata.uns["factor_sig"]["' + use_data+ '"]'+'["Factor_'+str(factor)+'"]["'+gene_sets+'"]["result"]')

    else:
        gl = gseapy.enrichr(gene_list=top_genes,
                            gene_sets=gene_sets,
                            cutoff=cutoff, outdir=None)

        result = gl.results

        result = result[result["Adjusted P-value"] < cutoff]
        
        
        adata.uns['enrichr_'+ gene_sets] = result


        print('Enrichment analysis of top genes is done! The result is stored in adata.uns["enrichr_'+gene_sets+ '""]')


    return adata if copy else None


import scipy as sp
from scipy import interpolate


def qvalue(pv, pi0=None):
    '''
    Estimates q-values from p-values

    This function is modified based on https://github.com/nfusi/qvalue

    Args
    ====
    pi0: if None, it's estimated as suggested in Storey and Tibshirani, 2003.

    '''
    assert(pv.min() >= 0 and pv.max() <= 1), "p-values should be between 0 and 1"

    original_shape = pv.shape
    pv = pv.ravel()  # flattens the array in place, more efficient than flatten()

    m = float(len(pv))

    # if the number of hypotheses is small, just set pi0 to 1
    if len(pv) < 100 and pi0 is None:
        pi0 = 1.0
    elif pi0 is not None:
        pi0 = pi0
    else:
        # evaluate pi0 for different lambdas
        pi0 = []
        lam = sp.arange(0, 0.90, 0.01)
        counts = sp.array([(pv > i).sum() for i in sp.arange(0, 0.9, 0.01)])
        for l in range(len(lam)):
            pi0.append(counts[l]/(m*(1-lam[l])))

        pi0 = sp.array(pi0)

        # fit natural cubic spline
        tck = interpolate.splrep(lam, pi0, k=3)
        pi0 = interpolate.splev(lam[-1], tck)

        if pi0 > 1:
            pi0 = 1.0

    assert(pi0 >= 0 and pi0 <= 1), "pi0 is not between 0 and 1: %f" % pi0

    p_ordered = sp.argsort(pv)
    pv = pv[p_ordered]
    qv = pi0 * m/len(pv) * pv
    qv[-1] = min(qv[-1], 1.0)

    for i in range(len(pv)-2, -1, -1):
        qv[i] = min(pi0*m*pv[i]/(i+1.0), qv[i+1])

    # reorder qvalues
    qv_temp = qv.copy()
    qv = sp.zeros_like(qv)
    qv[p_ordered] = qv_temp

    # reshape qvalues
    qv = qv.reshape(original_shape)

    return qv


from typing import Optional, Union
from anndata import AnnData
from .._compat import Literal
from PIL import Image
from pathlib import Path
# Test progress bar
from tqdm import tqdm
import numpy as np

def tiling(
        adata: AnnData,
        out_path: Union[Path, str] = "./",
        library_id: str = None,
        crop_size: int = 40,
        target_size: int = 299,
        verbose: bool = False,
        copy: bool = False
) -> Optional[AnnData]:
    """\
    Tiling H&E images to small tiles based on spot spatial location

    Parameters
    ----------
    adata
        Annotated data matrix.
    out_path
        Path to save spot image tiles
    library_id
        Library id stored in AnnData.
    crop_size
        Size of tiles
    verbose
        Verbose output
    copy
        Return a copy instead of writing to adata.
    target_size
        Input size for convolutional neuron network
    Returns
    -------
    Depending on `copy`, returns or updates `adata` with the following fields.
    **tile_path** : `adata.obs` field
        Saved path for each spot image tiles
    """

    if library_id is None:
        library_id = list(adata.uns["spatial"].keys())[0]

    image = (adata.uns["spatial"][library_id]["images"][adata.uns["spatial"]["use_quality"]]* 255).astype(np.uint8)

    img_pillow = Image.fromarray(image)
    tile_names = []

    with tqdm(total=len(adata), desc="Tiling image", bar_format="{l_bar}{bar} [ time left: {remaining} ]") as pbar:
        for imagerow, imagecol in zip(adata.obs["imagerow"], adata.obs["imagecol"]):
            imagerow_down = imagerow - crop_size / 2
            imagerow_up = imagerow + crop_size / 2
            imagecol_left = imagecol - crop_size / 2
            imagecol_right = imagecol + crop_size / 2
            tile = img_pillow.crop(
                (imagecol_left, imagerow_down, imagecol_right, imagerow_up))
            tile.thumbnail((target_size, target_size), Image.ANTIALIAS)
            tile.resize((target_size, target_size))
            tile_name = str(imagecol) + '-' + str(imagerow) + \
                '-' + str(crop_size)
            out_tile = Path(out_path) / (tile_name + '.jpeg')
            tile_names.append(str(out_tile))
            if verbose:
                print("generate tile at location ({}, {})".format(
                    str(imagecol), str(imagerow)))
            tile.save(out_tile, 'JPEG')

            pbar.update(1)

    adata.obs["tile_path"] = tile_names
    return adata if copy else None



from typing import Optional, Union
from anndata import AnnData
import pandas as pd
from pathlib import Path
import os


def positions(
    adata: AnnData,
    position_filepath: Union[Path, str] = None,
    scale_filepath: Union[Path, str] = None,
    quality: str = "low",
    copy: bool = False,
) -> Optional[AnnData]:

    """\
    Adding spatial information into the Anndata object

    Parameters
    ----------
    adata
        Annotated data matrix.
    position_filepath
        Path to tissue_positions_list file.
    scale_filepath
        Path to scalefactors_json file.
    quality
        Choosing low or high resolution image.
    copy
        Return a copy instead of writing to adata.
    Returns
    -------
    Depending on `copy`, returns or updates `adata` with the following fields.
    **imagecol** and **imagerow** : `adata.obs` field
        Spatial information of the tissue image.
    """

    tissue_positions = pd.read_csv(position_filepath, header=None)
    tissue_positions.columns = ["barcode", "tissue",
                                "row", "col", "imagerow", "imagecol"]
    import json
    with open(scale_filepath) as json_file:
        scale_factors = json.load(json_file)

    if quality == "low":
        tissue_positions["imagerow"] = tissue_positions["imagerow"] * \
            scale_factors['tissue_lowres_scalef']
        tissue_positions["imagecol"] = tissue_positions["imagecol"] * \
            scale_factors['tissue_lowres_scalef']
    elif quality == "high":
        tissue_positions["imagerow"] = tissue_positions["imagerow"] * \
            scale_factors['tissue_hires_scalef']
        tissue_positions["imagecol"] = tissue_positions["imagecol"] * \
            scale_factors['tissue_hires_scalef']

    tmp = adata.obs.merge(tissue_positions.reset_index()
                          .set_index(['barcode']),
                          left_index=True,
                          right_index=True,
                          how='left').reset_index()[["imagerow", "imagecol"]]

    adata.obs["imagerow"] = tmp["imagerow"].values
    adata.obs["imagecol"] = tmp["imagecol"].values

    return adata if copy else None



from typing import Optional, Union

import numpy as np
from anndata import AnnData
from numpy.random.mtrand import RandomState

from .._compat import Literal
import scanpy

_InitPos = Literal['paga', 'spectral', 'random']


def run_umap(
    adata: AnnData,
    min_dist: float = 0.5,
    spread: float = 1.0,
    n_components: int = 2,

    maxiter: Optional[int] = None,
    alpha: float = 1.0,
    gamma: float = 1.0,
    negative_sample_rate: int = 5,
    init_pos: Union[_InitPos, np.ndarray, None] = 'spectral',
    random_state: Optional[Union[int, RandomState]] = 0,
    a: Optional[float] = None,
    b: Optional[float] = None,
    copy: bool = False,
    method: Literal['umap', 'rapids'] = 'umap'
) -> Optional[AnnData]:
    """\
    Wrap function scanpy.pp.umap
    Embed the neighborhood graph using UMAP [McInnes18]_.
    UMAP (Uniform Manifold Approximation and Projection) is a manifold learning
    technique suitable for visualizing high-dimensional data. Besides tending to
    be faster than tSNE, it optimizes the embedding such that it best reflects
    the topology of the data, which we represent throughout Scanpy using a
    neighborhood graph. tSNE, by contrast, optimizes the distribution of
    nearest-neighbor distances in the embedding such that these best match the
    distribution of distances in the high-dimensional space.  We use the
    implementation of `umap-learn <https://github.com/lmcinnes/umap>`__
    [McInnes18]_. For a few comparisons of UMAP with tSNE, see this `preprint
    <https://doi.org/10.1101/298430>`__.
    Parameters
    ----------
    adata
        Annotated data matrix.
    n_components
        The number of dimensions of the embedding.
    random_state
        If `int`, `random_state` is the seed used by the random number generator;
        If `RandomState`, `random_state` is the random number generator;
        If `None`, the random number generator is the `RandomState` instance used
        by `np.random`.
    Returns
    -------
    Depending on `copy`, returns or updates `adata` with the following fields.
    `X_umap` : :class:`numpy.ndarray` (`adata.obsm`)
        Independent Component Analysis representation of data.

    """

    scanpy.tl.umap(adata, min_dist=min_dist, spread=spread, n_components=n_components,
         maxiter=maxiter, alpha=alpha, gamma=gamma, negative_sample_rate=negative_sample_rate,
         init_pos=init_pos, random_state=random_state, a=a, b=b, copy=copy, method=method)

    print(
        "UMAP is done! Generated in adata.obsm['X_umap'] nad adata.uns['umap']")




from types import MappingProxyType
from typing import Union, Optional, Any, Mapping, Callable

import numpy as np
import scipy
from anndata import AnnData
from numpy.random import RandomState
from .._compat import Literal
import scanpy

_Method = Literal['umap', 'gauss', 'rapids']
_MetricFn = Callable[[np.ndarray, np.ndarray], float]
# from sklearn.metrics.pairwise_distances.__doc__:
_MetricSparseCapable = Literal[
    'cityblock', 'cosine', 'euclidean', 'l1', 'l2', 'manhattan'
]
_MetricScipySpatial = Literal[
    'braycurtis', 'canberra', 'chebyshev', 'correlation', 'dice', 'hamming',
    'jaccard', 'kulsinski', 'mahalanobis', 'minkowski', 'rogerstanimoto',
    'russellrao', 'seuclidean', 'sokalmichener', 'sokalsneath', 'sqeuclidean',
    'yule'
]
_Metric = Union[_MetricSparseCapable, _MetricScipySpatial]


def neighbors(
    adata: AnnData,
    n_neighbors: int = 15,
    n_pcs: Optional[int] = None,
    use_rep: Optional[str] = None,
    knn: bool = True,
    random_state: Optional[Union[int, RandomState]] = 0,
    method: Optional[_Method] = 'umap',
    metric: Union[_Metric, _MetricFn] = 'euclidean',
    metric_kwds: Mapping[str, Any] = MappingProxyType({}),
    copy: bool = False,
) -> Optional[AnnData]:
    """\
    Compute a neighborhood graph of observations [McInnes18]_.
    The neighbor search efficiency of this heavily relies on UMAP [McInnes18]_,
    which also provides a method for estimating connectivities of data points -
    the connectivity of the manifold (`method=='umap'`). If `method=='gauss'`,
    connectivities are computed according to [Coifman05]_, in the adaption of
    [Haghverdi16]_.
    Parameters
    ----------
    adata
        Annotated data matrix.
    n_neighbors
        The size of local neighborhood (in terms of number of neighboring data
        points) used for manifold approximation. Larger values result in more
        global views of the manifold, while smaller values result in more local
        data being preserved. In general values should be in the range 2 to 100.
        If `knn` is `True`, number of nearest neighbors to be searched. If `knn`
        is `False`, a Gaussian kernel width is set to the distance of the
        `n_neighbors` neighbor.
    {n_pcs}
    {use_rep}
    knn
        If `True`, use a hard threshold to restrict the number of neighbors to
        `n_neighbors`, that is, consider a knn graph. Otherwise, use a Gaussian
        Kernel to assign low weights to neighbors more distant than the
        `n_neighbors` nearest neighbor.
    random_state
        A numpy random seed.
    method
        Use 'umap' [McInnes18]_ or 'gauss' (Gauss kernel following [Coifman05]_
        with adaptive width [Haghverdi16]_) for computing connectivities.
        Use 'rapids' for the RAPIDS implementation of UMAP (experimental, GPU
        only).
    metric
        A known metric’s name or a callable that returns a distance.
    metric_kwds
        Options for the metric.
    copy
        Return a copy instead of writing to adata.
    Returns
    -------
    Depending on `copy`, updates or returns `adata` with the following:
    **connectivities** : sparse matrix (`.uns['neighbors']`, dtype `float32`)
        Weighted adjacency matrix of the neighborhood graph of data
        points. Weights should be interpreted as connectivities.
    **distances** : sparse matrix (`.uns['neighbors']`, dtype `float32`)
        Instead of decaying weights, this stores distances for each pair of
        neighbors.
    """

    scanpy.pp.neighbors(adata, n_neighbors=n_neighbors, n_pcs=n_pcs, use_rep=use_rep,
              knn=knn, random_state=random_state, method=method, metric=metric,
              metric_kwds=metric_kwds, copy=copy)

    print("Created k-Nearest-Neighbor graph in adata.uns['neighbors'] ")



import numpy as np
import pandas as pd
from anndata import AnnData
from .het import create_windows


# Merge the L-R interaction results with cell type heterogeneity results to locate highly interacting sites

def merge(
    adata: AnnData,
    use_lr: str = 'lr_neighbours_louvain_max',
    use_het: pd.DataFrame = None,
) -> AnnData:
    """ Merge results from cell type heterogeneity and L-R clustering
    Parameters
    ----------
    adata: AnnData          The data object including the cell types to count
    use_lr: str             The largest expressed L-R cluster
    use_het: pd.DataFrame   The counts result of cell type heterogneity
                
    Returns
    -------
    counts: pd.DataFrame    The counts of specified clusters in each window of the tissue
    """
    coor = adata.obs[['imagerow', 'imagecol']]
    num = use_het.shape[0]
    windows, width, height = create_windows(adata, num)
    counts = pd.DataFrame(0, range(num), range(num))
    for n, window in enumerate(windows):
        spots = coor[(coor['imagecol'] > window[0]) & (coor['imagecol'] < window[0] + width) \
                    & (coor['imagerow'] < window[1]) & (coor['imagerow'] > window[1] - height)]
        try:
            counts.iloc[num-1-n%num, n//num] = sum(adata.obs[use_lr.split('_max')[0]][spots.index.tolist()] == adata.uns[use_lr]) / len(spots)
        except:
            counts.iloc[num-1-n%num, n//num] = 0
    # z-score for counts and use_het
    delta = counts.subtract(counts.mean().mean())
    std = (delta ** 2).sum().sum() / (delta.shape[0] * delta.shape[1])
    counts = delta.div(std)

    delta = use_het.subtract(use_het.mean().mean())
    std = (delta ** 2).sum().sum() / (delta.shape[0] * delta.shape[1])
    use_het = delta.div(std)

    adata.uns['merged'] = (counts + use_het).div(2)

    print("Results of spatial interaction analysis has been written to adata.uns['merged']")

    return adata



from typing import Union, Optional, Tuple, Collection, Sequence, Iterable
from anndata import AnnData
import numpy as np
from scipy.sparse import issparse, isspmatrix_csr, csr_matrix, spmatrix
import scanpy

def filter_genes(
    adata: AnnData,
    min_counts: Optional[int] = None,
    min_cells:  Optional[int] = None,
    max_counts: Optional[int] = None,
    max_cells:  Optional[int] = None,
    inplace: bool = True,
) -> Union[AnnData, None, Tuple[np.ndarray, np.ndarray]]:
    """\
    Wrap function scanpy.pp.filter_genes

    Filter genes based on number of cells or counts.
    Keep genes that have at least `min_counts` counts or are expressed in at
    least `min_cells` cells or have at most `max_counts` counts or are expressed
    in at most `max_cells` cells.
    Only provide one of the optional parameters `min_counts`, `min_cells`,
    `max_counts`, `max_cells` per call.
    Parameters
    ----------
    data
        An annotated data matrix of shape `n_obs` × `n_vars`. Rows correspond
        to cells and columns to genes.
    min_counts
        Minimum number of counts required for a gene to pass filtering.
    min_cells
        Minimum number of cells expressed required for a gene to pass filtering.
    max_counts
        Maximum number of counts required for a gene to pass filtering.
    max_cells
        Maximum number of cells expressed required for a gene to pass filtering.
    inplace
        Perform computation inplace or return result.
    Returns
    -------
    Depending on `inplace`, returns the following arrays or directly subsets
    and annotates the data matrix
    gene_subset
        Boolean index mask that does filtering. `True` means that the
        gene is kept. `False` means the gene is removed.
    number_per_gene
        Depending on what was tresholded (`counts` or `cells`), the array stores
        `n_counts` or `n_cells` per gene.
    """

    scanpy.pp.filter_genes(adata, min_counts=min_counts, min_cells=min_cells,
                 max_counts=max_counts, max_cells=max_cells, inplace=inplace)



from anndata import AnnData
from typing import Optional, Union
from scipy.stats import pearsonr
import gseapy
import pandas as pd


def microenv(
    adata: AnnData,
    use_data: str = "X_fa",
    factor: int = 1,
    gene_sets: str = "KEGG_2019_Human",
    cutoff: float = 0.05,
    n_top: int = 100,
    copy: bool = False,
) -> Optional[AnnData]:

    ft = adata.obsm[use_data][:,factor]

    genes = []
    result_pearson = []
    for gene in adata.var_names:
        
        genes.append(gene)
        result_pearson.append(pearsonr(ft,adata[:, gene].X.toarray().reshape(-1,len(adata))[0])[0])

    cor_results = pd.DataFrame(
        {'Gene': genes,
         'Pearson_correlation': result_pearson,
        })

    top_genes = list(cor_results.sort_values("Pearson_correlation",ascending=False)[:n_top].Gene)

    if "factor_sig" not in adata.uns:
            adata.uns['factor_sig'] = {}

    adata.uns['factor_sig'].update({use_data: {}})

    adata.uns['factor_sig'][use_data].update({"Factor_"+str(factor):{}})

    adata.uns['factor_sig'][use_data]["Factor_"+str(factor)].update({"top_genes" :top_genes}) 

    print('Get top genes of Factor ' + str(factor) + 
        ' is done! The top genes are stored in adata.uns["factor_sig"]["' + use_data+ '"]'+'["Factor_'+str(factor)+'"]["result"]')



    params = {'params':{'n_top': n_top,'factor': factor}}

    adata.uns['factor_sig'][use_data]["Factor_"+str(factor)].update({"params":params}) 



    return adata if copy else None


def checkType(arr, n=2):  
  
    # If the first two and the last two elements  
    # of the array are in increasing order  
    if (arr[0] <= arr[1] and 
        arr[n - 2] <= arr[n - 1]) : 
        return True  
  
    # If the first two and the last two elements  
    # of the array are in decreasing order  
    elif (arr[0] >= arr[1] and 
          arr[n - 2] >= arr[n - 1]) : 
        return False  


from matplotlib import pyplot as plt
from PIL import Image
import pandas as pd
import matplotlib
import numpy as np
import networkx as nx
import random
from stlearn._compat import Literal
from typing import Optional, Union
from anndata import AnnData
import warnings
import io
from copy import deepcopy


def tree_plot(
    adata: AnnData,
    library_id: str = None,
    figsize: Union[float,int] = (10,4),
    data_alpha: float = 1.0,
    use_label: str = "louvain",
    spot_size: Union[float,int] = 50,
    fontsize: int = 6,
    piesize: float = 0.15,
    zoom: float = 0.1,
    dpi: int = 180,
    show_all: bool = False,
    copy: bool = False,
) -> Optional[AnnData]:

    """\
    Hierarchical tree plot represent for the global spatial trajectory inference.

    Parameters
    ----------
    adata
        Annotated data matrix.
    library_id
        Library id stored in AnnData.
    use_label
        Use label result of clustering method.
    figsize
        Change figure size.
    data_alpha
        Opacity of the spot.
    fontsize
        Choose font size.
    piesize
        Choose the size of cropped image.
    zoom
        Choose zoom factor.
    show_all
        Show all cropped image or not.
    show_legend
        Show legend or not.
    dpi
        Set dpi as the resolution for the plot.
    copy
        Return a copy instead of writing to adata.
    Returns
    -------
    Nothing
    """


    G = adata.uns["PTS_graph"]

    if library_id is None:
        library_id = list(adata.uns["spatial"].keys())[0]

    for node in G.nodes:
        if node == 9999:
            break
        tmp_img = _generate_image(adata,library_id,sub_cluster=node,zoom=zoom, spot_size=spot_size,fontsize=fontsize,show_all=show_all)

        G.nodes[node]['image'] = tmp_img

    plt.rcParams['figure.dpi'] = dpi
    pos = hierarchy_pos(G,9999) 
    fig=plt.figure(figsize=figsize)
    a= plt.subplot(111)

    a.axis('off')
    nx.draw_networkx_edges(G,pos,a=a,arrowstyle="-",with_labels=True,edge_color="#ADABAF",connectionstyle="angle3,angleA=0,angleB=90")
    trans=a.transData.transform
    trans2=fig.transFigure.inverted().transform

    p2=piesize/2

    for n in G:
        if n == 9999:
            xx,yy=trans(pos[n]) # figure coordinates
            xa,ya=trans2((xx,yy)) # axes coordinates
            a = plt.axes([xa-p2,ya-p2, piesize, piesize])
            a.axis('off')
            a.text(0.5,0.9,"Pseudoroot",horizontalalignment='center',verticalalignment='center',
             transform=a.transAxes,bbox=dict(facecolor="#F9F9F9",boxstyle='round',edgecolor="#D1D1D1"))
            break
            
        xx,yy=trans(pos[n]) # figure coordinates
        xa,ya=trans2((xx,yy)) # axes coordinates
        a = plt.axes([xa-p2,ya-p2, piesize, piesize])

        subset = adata.obs[adata.obs["sub_cluster_labels"]==str(n)]
        color = adata.uns["tmp_color"][int(subset[use_label][0])]

        a.text(0.5,1.2,str(n),horizontalalignment='center',verticalalignment='center', 
            transform=a.transAxes,color='black',fontsize = fontsize,zorder=3,
               bbox=dict(facecolor=color,boxstyle='round'))
        #a.set_aspect('equal')
        a.axis('off')
        a.imshow(G.nodes[n]['image'])
        #plt.rcParams.update(plt.rcParamsDefault)




def _generate_image(adata,library_id,sub_cluster,zoom = 10,spot_size=100,fontsize=6,dpi=96,use_label="louvain",data_alpha=1,show_all=False,):
    
    #plt.rcParams['axes.linewidth'] = 4

    #plt.rcParams['axes.edgecolor'] = "#D1D1D1"

    subset = adata.obs[adata.obs["sub_cluster_labels"]==str(sub_cluster)]
    base = subset[["imagecol","imagerow"]].values
    if len(base)<25:
        zoom = zoom*20
        spot_size = spot_size*3
    elif len(base)>2000:
        zoom = zoom/4
        spot_size = spot_size/5
    x = base[:,0]
    y = base[:,1]
    #plt.rcParams['figure.dpi'] = 300

    fig2, ax2 = plt.subplots()

    for axis in ['top','bottom','left','right']:
        ax2.spines[axis].set_linewidth(0.5)
        ax2.spines[axis].set_color("#D1D1D1")

    color = adata.uns["tmp_color"][int(subset[use_label][0])]

    

    image = adata.uns["spatial"][library_id]["images"][adata.uns["spatial"]["use_quality"]]

    ax2.imshow(image,alpha=1)
    ax2.scatter(x,y,s=spot_size,alpha=data_alpha,edgecolor="none",c=color)
    
    #ax2.axis('off')
    ax2.get_xaxis().set_ticks([])
    ax2.get_yaxis().set_ticks([])

    try:
        ptp_bound = np.array(base).ptp(axis=0)
    except:
        print(base)
    
    m = 0
    n = 0
    if np.abs((y.min() - ptp_bound[1]*zoom) - (y.max() + ptp_bound[1]*zoom)) <50:
        m=m+50
    elif np.abs((x.min() - ptp_bound[1]*zoom) - (x.max() + ptp_bound[1]*zoom)) <50:
        n=n+50
    ax2.set_xlim(x.min() - ptp_bound[0]*zoom -n,
                x.max() + ptp_bound[0]*zoom -n)

    ax2.set_ylim(y.min() - ptp_bound[1]*zoom - m,
                y.max() + ptp_bound[1]*zoom + m)
    
    plt.gca().invert_yaxis()
    if show_all:
        plt.show()
    buf = io.BytesIO()
    fig2.savefig(buf, format='png', dpi = dpi,transparent=True,bbox_inches='tight',pad_inches=0.1)
    buf.seek(0)
    pil_img = deepcopy(Image.open(buf))
    plt.close(fig2)
    return pil_img






def hierarchy_pos(G, root=None, width=1., vert_gap = 0.2, vert_loc = 0, xcenter = 0.5):

    '''
    From Joel's answer at https://stackoverflow.com/a/29597209/2966723.  
    Licensed under Creative Commons Attribution-Share Alike 

    If the graph is a tree this will return the positions to plot this in a 
    hierarchical layout.

    G: the graph (must be a tree)

    root: the root node of current branch 
    - if the tree is directed and this is not given, 
      the root will be found and used
    - if the tree is directed and this is given, then 
      the positions will be just for the descendants of this node.
    - if the tree is undirected and not given, 
      then a random choice will be used.

    width: horizontal space allocated for this branch - avoids overlap with other branches

    vert_gap: gap between levels of hierarchy

    vert_loc: vertical location of root

    xcenter: horizontal location of root
    '''
    if not nx.is_tree(G):
        raise TypeError('cannot use hierarchy_pos on a graph that is not a tree')

    if root is None:
        if isinstance(G, nx.DiGraph):
            root = next(iter(nx.topological_sort(G)))  #allows back compatibility with nx version 1.11
        else:
            root = random.choice(list(G.nodes))

    def _hierarchy_pos(G, root, width=1., vert_gap = 0.2, vert_loc = 0, xcenter = 0.5, pos = None, parent = None):
        '''
        see hierarchy_pos docstring for most arguments

        pos: a dict saying where all nodes go if they have been assigned
        parent: parent of this branch. - only affects it if non-directed

        '''

        if pos is None:
            pos = {root:(xcenter,vert_loc)}
        else:
            pos[root] = (xcenter, vert_loc)
        children = list(G.neighbors(root))
        if not isinstance(G, nx.DiGraph) and parent is not None:
            children.remove(parent)  
        if len(children)!=0:
            dx = width/len(children) 
            nextx = xcenter - width/2 - dx/2
            for child in children:
                nextx += dx
                pos = _hierarchy_pos(G,child, width = dx, vert_gap = vert_gap, 
                                    vert_loc = vert_loc-vert_gap, xcenter=nextx,
                                    pos=pos, parent = root)
        return pos


    return _hierarchy_pos(G, root, width, vert_gap, vert_loc, xcenter)


from matplotlib import pyplot as plt
from PIL import Image
import pandas as pd
import matplotlib
import numpy as np
import networkx as nx
from stlearn._compat import Literal
from typing import Optional, Union
from anndata import AnnData
import warnings
#from .utils import get_img_from_fig, checkType

def pseudotime_plot(
    adata: AnnData,
    library_id: str = None,
    use_label: str = "louvain",
    list_cluster: Union[str,list] = None,
    data_alpha: float = 1.0,
    tissue_alpha: float = 1.0,
    edge_alpha: float = 0.8,
    node_alpha: float = 1.0,
    spot_size: Union[float,int] = 6.5,
    node_size: float = 5,
    show_color_bar: bool = True,
    show_axis: bool = False,
    show_graph: bool = True,
    dpi: int = 180,
    copy: bool = False,
) -> Optional[AnnData]:

    """\
    Global trajectory inference plot (Only DPT).

    Parameters
    ----------
    adata
        Annotated data matrix.
    library_id
        Library id stored in AnnData.
    use_label
        Use label result of clustering method.
    list_cluster
        Choose set of clusters that will display in the plot.
    data_alpha
        Opacity of the spot.
    tissue_alpha
        Opacity of the tissue.
    edge_alpha
        Opacity of edge in PAGA graph in the tissue.
    node_alpha
        Opacity of node in PAGA graph in the tissue.
    cmap
        Color map to use.
    spot_size
        Size of the spot.
    node_size
        Size of node in PAGA graph in the tissue.
    show_color_bar
        Show color bar or not.
    show_axis
        Show axis or not.
    show_graph
        Show PAGA graph or not.
    show_legend
        Show legend or not.
    dpi
        Set dpi as the resolution for the plot.
    show_trajectory
        Show the spatial trajectory or not. It requires stlearn.spatial.trajectory.pseudotimespace.
    show_subcluster
        Show subcluster or not. It requires stlearn.spatial.trajectory.global_level.
    copy
        Return a copy instead of writing to adata.
    Returns
    -------
    Nothing
    """
    
    plt.rcParams['figure.dpi'] = dpi

    if list_cluster == "all":
        list_cluster = list(range(0,len(adata.obs[use_label].unique())))
    # Get query clusters
    command = []
    for i in list_cluster:
        command.append(use_label + ' == "' + str(i) + '"')
    tmp = adata.obs.query(" or ".join(command))

    G=adata.uns["global_graph"]

    labels = nx.get_edge_attributes(G,'weight')

    result = []
    query_node = get_node(list_cluster,adata.uns["split_node"])
    for edge in G.edges(query_node):
        if (edge[0] in query_node) and (edge[1] in query_node):
            result.append(edge)

    result2 = []
    for edge in result:
        try:
            result2.append(labels[edge]+0.5)
        except: 
            result2.append(labels[edge[::-1]]+0.5)

    fig, a = plt.subplots()
    centroid_dict = adata.uns["centroid_dict"]
    dpt = adata.obs["dpt_pseudotime"]

    colors = adata.obs[use_label].astype(int)
    vmin = min(dpt)
    vmax = max(dpt)
    # Plot scatter plot based on pixel of spots
    from sklearn.preprocessing import MinMaxScaler
    scaler = MinMaxScaler()
    scale = scaler.fit_transform(tmp['dpt_pseudotime'].values.reshape(-1,1)).reshape(-1,1)

    plot = a.scatter(tmp["imagecol"], tmp["imagerow"], edgecolor="none", alpha=data_alpha,s=spot_size,marker="o",
               vmin=vmin, vmax=vmax,cmap=plt.get_cmap("viridis"),c=scale.reshape(1,-1)[0])

    n_clus = len(colors.unique())
    
    from scanpy.plotting import palettes
    cmaps = matplotlib.colors.LinearSegmentedColormap.from_list("",palettes.vega_20_scanpy)

    cmap = plt.get_cmap(cmaps)
    bounds=np.linspace(0, n_clus, n_clus+1)
    norm = matplotlib.colors.BoundaryNorm(bounds, cmap.N)
    


    norm = matplotlib.colors.Normalize(vmin=min(colors), vmax=max(colors))
    m = matplotlib.cm.ScalarMappable(norm=norm, cmap=cmap)

    if show_graph:
        nx.draw_networkx_edges(G, pos=centroid_dict,node_size=1,font_size=0,linewidths=2,edgelist=result,
                                width=result2,alpha=edge_alpha,edge_color='#333333',)

        for x,y in centroid_dict.items():

            if(x in get_node(list_cluster,adata.uns["split_node"])):
                a.text(y[0],y[1],get_cluster(str(x),adata.uns["split_node"]),color='white',fontsize = node_size,zorder=100,
                       bbox=dict(facecolor=cmap(int(get_cluster(str(x),adata.uns["split_node"]))/19),boxstyle='circle',alpha=node_alpha))
    if show_color_bar:
        cb = plt.colorbar(plot,cmap="viridis")
        cb.outline.set_visible(False)

    if library_id is None:
        library_id = list(adata.uns["spatial"].keys())[0]

    image = adata.uns["spatial"][library_id]["images"][adata.uns["spatial"]["use_quality"]]
    
    a.imshow(image,alpha=tissue_alpha, zorder=-1,)

    if not show_axis:
        a.axis('off')
    a.set_title
    plt.show()

# get name of cluster by subcluster
def get_cluster(search,dictionary):
    for cl, sub in dictionary.items():    # for name, age in dictionary.iteritems():  (for Python 2.x)
        if search in sub:
            return(cl)
        
def get_node(node_list,split_node):
    result = np.array([])
    for node in node_list:
        result = np.append(result,np.array(split_node[node]).astype(int))
    return result.astype(int)


"""Reading and Writing
"""
from pathlib import Path, PurePath
from typing import Union

from anndata import AnnData


def file_10x_h5(
    filename: Union[str, Path],
) -> AnnData:

    """\
    Wrap function scanpy.read_10x_h5
    Read 10x-Genomics-formatted hdf5 file.

    Parameters
    ----------
    filename
        Path to the h5 file.
    Returns
    -------
    AnnData
    """

    from scanpy.readwrite import read_10x_h5

    adata = read_10x_h5(filename)

    return adata


def file_10x_mtx(
    filename: Union[str, Path],
) -> AnnData:
    """\
    Wrap function scanpy.read_10x_mtx
    Read .mtx file.

    Parameters
    ----------
    filename
        Path to the h5 file.
    Returns
    -------
    AnnData
    """

    from scanpy.readwrite import read_10x_mtx

    adata = read_10x_mtx(filename)

    return adata


def file_table(
    filename: Union[str, Path],
) -> AnnData:
    """
    Wrap function scanpy.read
    Read file and return AnnData object.

    Parameters
    ----------
    filename
        Path to the h5 file.
    Returns
    -------
    AnnData
    """

    from scanpy.readwrite import read

    adata = read(filename)

    return adata



import numpy as np
import pandas as pd
from anndata import AnnData

def create_windows(
    adata: AnnData,
    num: int,   
):
    """ Generate screening windows across the tissue sample
    Parameters
    ----------
    adata: AnnData          The data object to generate windows on
    num: int                Number of windows on X- and Y- axes
    Returns
    -------
    windows                 The individual windows defined by left and upper side
    width                   Width of windows
    height                  Height of windows
    """   
    coor = adata.obs[['imagerow', 'imagecol']]
    max_x = max(coor['imagecol'])
    min_x = min(coor['imagecol'])
    max_y = max(coor['imagerow'])
    min_y = min(coor['imagerow'])
    width = (max_x - min_x) / num
    height = (max_y - min_y) / num
    windows = []
    # prepare heatmap data (top down and left to right)
    for i in range(num * num):
        x = min_x + i // num * width  # left side
        y = max_y - i % num * height  # upper side
        windows.append([x, y])

    return windows, width, height


def count(
    adata: AnnData,
    num: int = 10,
    use_clustering: str = 'louvain',
    use_clusters: list = [],
) -> pd.DataFrame:
    """ Count the cell type densities
    Parameters
    ----------
    adata: AnnData          The data object including the cell types to count
    num: int                Number of windows on X- and Y-axes
    use_clustering:         The cell type results to use in counting  
    use_clusters:           Specify certain cluster(s) in counting
     
    Returns
    -------
    counts: pd.DataFrame    The counts of specified clusters in each window of the tissue
    """
    if len(use_clusters) == 0:
        use_clusters = list(set(adata.obs[use_clustering]))
    coor = adata.obs[['imagerow', 'imagecol']]
    cluster = adata.obs[use_clustering]
    windows, width, height = create_windows(adata, num)
    counts = pd.DataFrame(0, range(num), range(num))
    for n, window in enumerate(windows):
        spots = coor[(coor['imagecol'] > window[0]) & (coor['imagecol'] < window[0] + width) \
                    & (coor['imagerow'] < window[1]) & (coor['imagerow'] > window[1] - height)]
        counts.iloc[num-1-n%num, n//num] = len(set([i for i in cluster[spots.index.tolist()] if i in use_clusters]))

    return counts



from .plotting.gene_plot import gene_plot
from .plotting.cluster_plot import cluster_plot
from .plotting.subcluster_plot import subcluster_plot
from .plotting.microenv_plot import microenv_plot
from .plotting.non_spatial_plot import non_spatial_plot
from .plotting import trajectory
from .plotting.QC_plot import QC_plot
from .plotting.cci_plot import het_plot
from .plotting.cci_plot import violin_plot
from .plotting.cci_plot import stacked_bar_plot



from PIL import Image
import numpy as np
import matplotlib.pyplot as plt
from typing import Optional, Union
from anndata import AnnData
#import cv2
import io
from matplotlib import cm


def enrichment_analysis_plot(
    adata: AnnData,
    method: str = "fa",
    factor: int = None,
    gene_sets: str = None,
    save_plot: str = None,
    cmap: str = "plasma",
    title: str = None,
    dpi: int = 192,
    store: bool = True,
    output: str = None,
    copy: bool = False,
) -> Optional[AnnData]:

    plt.rcParams['figure.dpi'] = dpi
    tmp = adata.uns["factor_sig"][method]["Factor_" +
                                          str(factor)][gene_sets]["result"]
    objects = tmp.Term[::-1]

    y_pos = np.arange(len(objects))
    performance = tmp['Odds Ratio'][::-1]
    fig, a = plt.subplots()

    y_pvalue = tmp["Adjusted P-value"][::-1]

    plot = plt.scatter(y_pvalue, y_pvalue, c=y_pvalue, cmap='plasma')

    colors = cm.plasma_r(y_pos / float(max(y_pos)))
    plt.clf()
    plt.colorbar(plot, orientation="horizontal", pad=0.2)
    plt.barh(range(len(y_pos)), performance, color=colors)

    #a.barh(y_pos, performance, align='center', alpha=0.5)
    plt.yticks(y_pos, objects)
    plt.xlabel('Odd ratio \n Adjusted P-value')
    if title is None:
        plt.title(gene_sets)
    else:
        plt.title(title)

    if output is not None:
        fig.savefig(output + "/enrich_plot_factor_" + str(factor) + "_" +
                    gene_sets + ".png", dpi=dpi, bbox_inches='tight', pad_inches=0)

    if store:

        fig_np = get_img_from_fig(fig, dpi)

        plt.close(fig)

        current_plot = {"img": fig_np}

        adata.uns["factor_sig"][method]["Factor_" +
                                        str(factor)][gene_sets].update(current_plot)

    print('The plot stored in adata.uns["factor_sig"]["' + method +
          '"]'+'["Factor_'+str(factor)+'"]["'+gene_sets+'"]["img"]')


def get_img_from_fig(fig, dpi=180):
    buf = io.BytesIO()
    from io import BytesIO

    fig.savefig(buf, format="png", dpi=dpi, bbox_inches='tight', pad_inches=0)
    buf.seek(0)
    img_arr = np.frombuffer(buf.getvalue(), dtype=np.uint8)
    img = np.asarray(Image.open(BytesIO(img_arr)))
    buf.close()
    #img = cv2.imdecode(img_arr, 1)
    #img = cv2.cvtColor(img, cv2.COLOR_BGRA2RGB)

    return img



import numpy as np
import pandas as pd
from typing import Optional

from anndata import AnnData
from sklearn.decomposition import FactorAnalysis
from scipy.sparse import issparse


def run_fa(
    adata: AnnData,
    n_factors: int = 20,
    tol: float = 0.01,
    max_iter: int = 1000,
    svd_method: str = "randomized",
    iterated_power: int = 3,
    random_state: int = 2108,
    use_data: str = None,
    copy: bool = False,
) -> Optional[AnnData]:

    """\
    Factor Analysis (FA)
    A simple linear generative model with Gaussian latent variables.
    The observations are assumed to be caused by a linear transformation of
    lower dimensional latent factors and added Gaussian noise.
    Without loss of generality the factors are distributed according to a
    Gaussian with zero mean and unit covariance. The noise is also zero mean
    and has an arbitrary diagonal covariance matrix.
    If we would restrict the model further, by assuming that the Gaussian
    noise is even isotropic (all diagonal entries are the same) we would obtain
    :class:`PPCA`.
    FactorAnalysis performs a maximum likelihood estimate of the so-called
    `loading` matrix, the transformation of the latent variables to the
    observed ones, using SVD based approach.

    Parameters
    ----------
    adata
        Annotated data matrix.
    n_components
        Dimensionality of latent space, the number of components
        of ``X`` that are obtained after ``transform``.
        If None, n_components is set to the number of features.
    tol
        Stopping tolerance for log-likelihood increase.
    max_iter
        Maximum number of iterations.
    noise_variance_init
        The initial guess of the noise variance for each feature.
        If None, it defaults to np.ones(n_features)
    svd_method
        Which SVD method to use. If 'lapack' use standard SVD from
        scipy.linalg, if 'randomized' use fast ``randomized_svd`` function.
        Defaults to 'randomized'. For most applications 'randomized' will
        be sufficiently precise while providing significant speed gains.
        Accuracy can also be improved by setting higher values for
        `iterated_power`. If this is not sufficient, for maximum precision
        you should choose 'lapack'.
    iterated_power
        Number of iterations for the power method. 3 by default. Only used
        if ``svd_method`` equals 'randomized'
    random_state
        If int, random_state is the seed used by the random number generator;
        If RandomState instance, random_state is the random number generator;
        If None, the random number generator is the RandomState instance used
        by `np.random`. Only used when ``svd_method`` equals 'randomized'.
    copy
        Return a copy instead of writing to adata.
    Returns
    -------
    Depending on `copy`, returns or updates `adata` with the following fields.
    `X_fa` : :class:`numpy.ndarray` (`adata.obsm`)
        Factor analysis representation of data.
    """

    if use_data is None:
        if issparse(adata.X):
            matrix = adata.X.toarray()
        else:
            matrix = adata.X

    else:
        matrix = adata.obsm[use_data].values

    fa = FactorAnalysis(n_components=n_factors, tol=tol, max_iter=max_iter,
                        svd_method=svd_method, iterated_power=iterated_power,
                        random_state=random_state)

    latent = fa.fit_transform(matrix)

    adata.obsm["X_fa"] = latent

    adata.uns['fa_params'] = {'params': {'n_factors': n_factors, 'tol': tol,
                                         'max_iter': max_iter, 'svd_method': svd_method,
                                         'iterated_power': iterated_power,
                                         'random_state': random_state}}

    print('FA is done! Generated in adata.obsm["X_fa"]')

    return adata if copy else None



from sklearn.cluster import KMeans
from anndata import AnnData
from typing import Optional, Union
import pandas as pd
import numpy as np
from natsort import natsorted

def kmeans(
    adata: AnnData,
    n_clusters: int = 20,
    use_data: str = "X_pca",
    init: str = "k-means++",
    n_init: int = 10,
    max_iter: int = 300,
    tol: float = 0.0001,
    precompute_distances: str = "auto",
    random_state: str = None,
    copy_x: bool = True,
    n_jobs: int = None, 
    algorithm: str = "auto",
    key_added: str = "kmeans",
    copy: bool = False,
) -> Optional[AnnData]:

    """\
    Perform kmeans clustering for spatial transcriptomics data

    Parameters
    ----------
    adata
        Annotated data matrix.
    n_clusters 
        The number of clusters to form as well as the number of
        centroids to generate.
    use_data
        Use dimensionality reduction result.
    init
        Method for initialization, defaults to 'k-means++'
    max_iter
        Maximum number of iterations of the k-means algorithm for a
        single run.
    tol
        Relative tolerance with regards to inertia to declare convergence.
    precompute_distances
        Precompute distances (faster but takes more memory).
    random_state
        Determines random number generation for centroid initialization. Use
        an int to make the randomness deterministic.
    copy_x
        When pre-computing distances it is more numerically accurate to center
        the data first.  If copy_x is True (default), then the original data is
        not modified, ensuring X is C-contiguous.  If False, the original data
        is modified, and put back before the function returns, but small
        numerical differences may be introduced by subtracting and then adding
        the data mean, in this case it will also not ensure that data is
        C-contiguous which may cause a significant slowdown.
    n_jobs
        The number of jobs to use for the computation. This works by computing
        each of the n_init runs in parallel.
    algorithm
        K-means algorithm to use. The classical EM-style algorithm is "full".
        The "elkan" variation is more efficient by using the triangle
        inequality, but currently doesn't support sparse data. "auto" chooses
        "elkan" for dense data and "full" for sparse data.
    key_added
        Key add to adata.obs
    copy
        Return a copy instead of writing to adata.
    Returns
    -------
    Anndata
    """
    
    data = adata.obsm[use_data]
    
    print("Applying Kmeans clustering ...")
        
    kmeans = KMeans(n_clusters=n_clusters,
                    init = init, 
                    n_init=n_init,
                    max_iter=max_iter,
                    tol=tol, 
                    precompute_distances=precompute_distances,
                    random_state=random_state,
                    copy_x=copy_x,
                    n_jobs=n_jobs,
                    algorithm=algorithm).fit(data)

    adata.obs[key_added] = pd.Categorical(
        values=np.array(kmeans.labels_).astype('U'),
        categories=natsorted(np.unique(np.array(kmeans.labels_)).astype('U')),
    )

    print('Kmeans clustering is done! The labels are stored in adata.obs["kmeans"]')
    
    return adata if copy else None


from typing import Optional, Union, Iterable, Dict

import numpy as np
from anndata import AnnData
from scipy.sparse import issparse
from sklearn.utils import sparsefuncs
from stlearn._compat import Literal
import scanpy

def normalize_total(
    adata: AnnData,
    target_sum: Optional[float] = None,
    exclude_highly_expressed: bool = False,
    max_fraction: float = 0.05,
    key_added: Optional[str] = None,
    layers: Union[Literal['all'], Iterable[str]] = None,
    layer_norm: Optional[str] = None,
    inplace: bool = True,
) -> Optional[Dict[str, np.ndarray]]:
    """\
    Wrap function from scanpy.pp.log1p
    Normalize counts per cell.
    If choosing `target_sum=1e6`, this is CPM normalization.
    If `exclude_highly_expressed=True`, very highly expressed genes are excluded
    from the computation of the normalization factor (size factor) for each
    cell. This is meaningful as these can strongly influence the resulting
    normalized values for all other genes [Weinreb17]_.
    Similar functions are used, for example, by Seurat [Satija15]_, Cell Ranger
    [Zheng17]_ or SPRING [Weinreb17]_.
    Params
    ------
    adata
        The annotated data matrix of shape `n_obs` × `n_vars`.
        Rows correspond to cells and columns to genes.
    target_sum
        If `None`, after normalization, each observation (cell) has a total
        count equal to the median of total counts for observations (cells)
        before normalization.
    exclude_highly_expressed
        Exclude (very) highly expressed genes for the computation of the
        normalization factor (size factor) for each cell. A gene is considered
        highly expressed, if it has more than `max_fraction` of the total counts
        in at least one cell. The not-excluded genes will sum up to
        `target_sum`.
    max_fraction
        If `exclude_highly_expressed=True`, consider cells as highly expressed
        that have more counts than `max_fraction` of the original total counts
        in at least one cell.
    key_added
        Name of the field in `adata.obs` where the normalization factor is
        stored.
    layers
        List of layers to normalize. Set to `'all'` to normalize all layers.
    layer_norm
        Specifies how to normalize layers:
        * If `None`, after normalization, for each layer in *layers* each cell
          has a total count equal to the median of the *counts_per_cell* before
          normalization of the layer.
        * If `'after'`, for each layer in *layers* each cell has
          a total count equal to `target_sum`.
        * If `'X'`, for each layer in *layers* each cell has a total count
          equal to the median of total counts for observations (cells) of
          `adata.X` before normalization.
    inplace
        Whether to update `adata` or return dictionary with normalized copies of
        `adata.X` and `adata.layers`.
    Returns
    -------
    Returns dictionary with normalized copies of `adata.X` and `adata.layers`
    or updates `adata` with normalized version of the original
    `adata.X` and `adata.layers`, depending on `inplace`.
    """


    scanpy.pp.normalize_total(adata, target_sum=target_sum,
                    exclude_highly_expressed=exclude_highly_expressed,
                    max_fraction=max_fraction, key_added=key_added,
                    layers=layers, layer_norm=layer_norm, inplace=inplace)

    print("Normalization step is finished in adata.X")



from itertools import cycle
from PIL import ImageTk, Image
import sys
if sys.version_info[0] < 3:
    import Tkinter as tk     ## Python 2.x
else:
    import tkinter as tk     ## Python 3.x
from tkinter.filedialog import asksaveasfile 
from tkinter import messagebox,Menu
from anndata import AnnData
from stlearn._compat import Literal
from typing import Optional

import matplotlib.pyplot as plt

class ChangeImage():
    def __init__(self, root, adata, method):
        self.adata = adata
        self.method = method
        self.pil_imgs = []
        self.photos=[]
        self.routes = []
        self.load_images(root)
        self.image_num=0
        self.canvas = tk.Canvas(root, width=self.pil_imgs[0].size[0], 
                                height=self.pil_imgs[0].size[1],bg="#fff",bd=0)
        self.canvas.pack(fill=tk.BOTH,expand=True,side=tk.TOP)  
        self.canvas.bind('<Configure>', self._resize_image)
        self.image_on_canvas = self.canvas.create_image(0, 0, 
                    anchor=tk.NW, image=self.photos[self.image_num])

        ## click on button to change image
        self.bt_plot = tk.Button(root, text="Plot!", 
                                command=lambda:[plot(self.adata,
                                S=int(self.routes[self.image_num][1])),root.destroy()],
                                  bg="#007bff",activebackground="#0053ac",foreground="#fff",
                                  activeforeground="#fff",bd=0)
        self.bt_plot.pack(side=tk.RIGHT, padx=5, expand=True)

        menu = Menu(root,bg="#fff",bd=0)
        root.config(menu=menu,width = 20,bg="#fff",bd=0)
        filemenu = Menu(menu)
        menu.add_cascade(label="  File  ", menu=filemenu)
        filemenu.add_command(label="  Save  ", command=self.save)
        filemenu.add_separator()
        filemenu.add_command(label="  Exit  ", command=root.destroy)


        self.bt_next = tk.Button(root, text="Next", command=self.next_image,
                                  bg="#007bff",activebackground="#0053ac",foreground="#fff",
                                  activeforeground="#fff",bd=0)
        self.bt_next.pack(side=tk.RIGHT,expand=True)
        self.bt_next.config( height = 1, width = 5 )
        self.bt_prev = tk.Button(root, text="Prev", command=self.previous_image,
                                  bg="#007bff",activebackground="#0053ac",foreground="#fff",
                                  activeforeground="#fff",bd=0)
        self.bt_prev.pack(side=tk.LEFT,expand=True)
        self.bt_prev.config( height = 1, width = 5 )
        self.bt_select = tk.Button(root, text="Select", command=self.select_images,
                                  bg="#007bff",activebackground="#0053ac",foreground="#fff",
                                  activeforeground="#fff",bd=0)
        self.bt_select.pack(fill=tk.X, padx=5, expand=True)
        self.bt_select.config( height = 1, width = 5 )
        self.frame1 = tk.Frame(root,bg="#fff")
        self.frame1.pack(fill=tk.X)
        self.lbl1 = tk.Label(self.frame1, text=("Root node: "), width=15,bg="#fff")
        self.lbl1.pack(side=tk.LEFT, padx=5, pady=5)
        self.v = tk.StringVar()
        self.v.set(self.adata.uns["pseudotimespace"]["routes"][self.routes[self.image_num]])

        self.lbl2 = tk.Label(self.frame1, text=(self.routes[self.image_num]), width=15,
                            bg="#fff",bd=0)
        self.lbl2.pack(side=tk.RIGHT, padx=5, pady=5)

        self.entry1 = tk.Entry(self.frame1,textvariable=self.v,bg="#dedede",bd=0)
        self.entry1.bind('<Return>', self.select_images)
        self.entry1.pack(fill=tk.X, padx=5, expand=True)
        #self.bt_exit = tk.Button(root, text="Exit", bg="orange", command=root.destroy)
        #self.bt_exit.pack(side=tk.LEFT)
    def next_image(self):
        self.image_num += 1
        if self.image_num >= len(self.photos):
            self.image_num=0
        ## pipe the next image to be displayed to the button
        self.image = self.pil_imgs[self.image_num].resize((self.new_width, self.new_height))
        self.image = ImageTk.PhotoImage(self.image)
        
        self.canvas.itemconfig(self.image_on_canvas, image =self.image)
        self.v.set(self.adata.uns["pseudotimespace"]["routes"][self.routes[self.image_num]])
        
        self.lbl2['text'] = self.routes[self.image_num]
    def previous_image(self):
        self.image_num = self.image_num - 1
        if self.image_num < 0:
            self.image_num=len(self.photos)-1
        ## pipe the next image to be displayed to the button
        self.image = self.pil_imgs[self.image_num].resize((self.new_width, self.new_height))
        self.image = ImageTk.PhotoImage(self.image)
        self.canvas.itemconfig(self.image_on_canvas, image =self.image)
        self.v.set(self.adata.uns["pseudotimespace"]["routes"][self.routes[self.image_num]])
        
        self.lbl2['text'] = self.routes[self.image_num]
        
    def load_images(self,root):
        """ copy data images to a list that is an instance variable
            all images are hard-coded here and so use data=
            instead of file=
        """
        ## put the images in an instance object (self.) so they aren't destroyed
        ## when the function exits and can be used anywhere in the class
        for item in self.adata.uns['plots']['trajectories'].items():
            self.pil_imgs.append(Image.fromarray(item[1]))
            self.photos.append(ImageTk.PhotoImage(image=Image.fromarray(item[1]),master=root))
            self.routes.append(item[0])
    def select_images(self,*args):
        if ((len(self.v.get()) == 0) or (int(self.v.get()) not in self.adata.uns['pseudotimespace']['flat_tree'].nodes)):
            messagebox.showerror("Error","Please choose available root! (" +
                     str(self.adata.uns['pseudotimespace']['flat_tree'].nodes) + ")")
            return None


        self.image_num = int(self.v.get())-1
        self.image = self.pil_imgs[self.image_num].resize((self.new_width, self.new_height))
        self.image = ImageTk.PhotoImage(self.image)
        self.canvas.itemconfig(self.image_on_canvas, image =self.image)
    
    def save(self): 
        f = asksaveasfile(filetypes=(("Portable Network Graphics (*.png)", "*.png"),
                                        ("All Files (*.*)", "*.*")),
                             mode='wb',
                             defaultextension='.png')
        if f is None:
            return

        filename = f.name
        extension = filename.rsplit('.', 1)[-1]
        img = Image.fromarray(self.adata.uns['plots']["trajectories"][self.routes[self.image_num]])
        img.save(f, extension)
        f.close()
        
    def _resize_image(self,event):

        self.new_width = event.width
        self.new_height = event.height
        

        self.image = self.pil_imgs[self.image_num].resize((self.new_width, self.new_height))
        self.image = ImageTk.PhotoImage(self.image)
        self.canvas.itemconfig(self.image_on_canvas, image =self.image)
        

def pseudotimespace_plot(
    adata: AnnData,
    method: str = "trajectories",
    root: int = None,
    S: int = None,
    ):
    
    if ((S is None) and (root is None)):
    
        root=tk.Tk()
        root.title('stLearn - Trajectories Inferences - Pseudo-time-space')
        CI=ChangeImage(root,adata,method)
        root.mainloop()

    elif ((S is not None) and (root is not None)):
        raise ValueError(
            'Only provide one of the optional parameters "S" or "root" ')
    else:
        pts_plot(adata,root=root,S=S)



def pts_plot(
    adata: AnnData,
    method: str = "trajectories",
    dpi: int = 192,):
    
    if method not in ["trajectories"]:
        raise ValueError("Please provide one of these methods : fa, ica or ldvae")

    if method not in str(adata.obsm):
        raise ValueError("Please run trajectories pipeline")
        

    root=tk.Tk()
    root.title('stLearn - Pseudo-time-space - Trajectories')
    CI=ChangeImage(root,adata,method)
    root.mainloop()



def plot(
    adata: AnnData,
    method: str = "trajectories",
    root: int = None,
    S: int = None,
    dpi: int = 192,
    copy: bool = False,
) -> Optional[AnnData]:
    

    if ((S is not None) and (root is not None)):
        raise ValueError(
            'Only provide one of the optional parameters "S" or "root" ')

    plt.rcParams['figure.dpi'] = dpi

    if S is not None:
        
        plt.imshow(adata.uns["plots"]["trajectories"]["S" + str(S) + "_pseudotime"])
        plt.axis('off')
        plt.show()
    else:
        dictionary = adata.uns["pseudotimespace"]["routes"]

        route = find_routes(root,dictionary)

        plt.axis('off')
        plt.imshow(adata.uns["plots"]["trajectories"][route])
        
        plt.show()





def find_routes(pattern,dictionary):
    
    for route, root in dictionary.items():
        if root == pattern:
            return route


from typing import Optional, Union
from anndata import AnnData
from matplotlib import pyplot as plt
from pathlib import Path
import os


def annotation(
    adata: AnnData,
    label_list: list,
    use_label: str = "louvain",
    copy: bool = False,
) -> Optional[AnnData]:

    """\
    Adding annotation for cluster

    Parameters
    ----------
    adata
        Annotated data matrix.
    label_list
        List of the labels which assigned to current clustering result.
    use_label
        Choosing clustering type.
    copy
        Return a copy instead of writing to adata.
    Returns
    -------
    Depending on `copy`, returns or updates `adata` with the following fields.
    **[clustering method name]_anno** : `adata.obs` field
        The annotation of cluster results.
    """

    if label_list is None:
        raise ValueError("Please give the label list!")

    if len(label_list) != len(adata.obs[use_label].unique()):
        raise ValueError("Please give the correct number of label list!")

    adata.obs[use_label + '_anno'] = adata.obs[use_label]
    adata.obs[use_label + '_anno'].cat.categories = label_list

    print("The annotation is added to adata.obs['" + use_label + "_anno"+"']")

    return adata if copy else None



from anndata import AnnData
import stlearn
import os
import scanpy

def Preprocessing(
    adata: AnnData,
    out_path: str = "../tiling",
    crop_size: float = 40,
    use_data: str = "X_pca",
    radius: float = 50,
    adjust_method: str = "mean",
    n_neighbors: int = 25,
    use_label: str = "louvain",
    n_clusters: int = 10,
    random_state: int = 0,
    copy: bool = False,

    ) -> AnnData:

    scanpy.pp.filter_genes()
    scanpy.pp.filter_cells()
    scanpy.pp.normalize_per_cell(adata, counts_per_cell_after=1e6)
    scanpy.pp.log1p(adata)

    return adata if copy else None



import numpy as np
import pandas as pd
from typing import Optional
from anndata import AnnData
from sklearn.decomposition import FastICA
from scipy.sparse import issparse


def run_ica(
    adata: AnnData,
    n_factors: int = 20,
    fun: str = "logcosh",
    tol: float = 0.0001,
    use_data: str = None,
    copy: bool = False,
) -> Optional[AnnData]:

    """\
    FastICA: a fast algorithm for Independent Component Analysis.

    Parameters
    ----------
    adata
        Annotated data matrix.
    n_factors
        Number of components to use. If none is passed, all are used.
    fun
        The functional form of the G function used in the
        approximation to neg-entropy. Could be either 'logcosh', 'exp',
        or 'cube'.
        You can also provide your own function. It should return a tuple
        containing the value of the function, and of its derivative, in the
        point. Example:
        def my_g(x):
            return x ** 3, (3 * x ** 2).mean(axis=-1)
    tol
        Tolerance on update at each iteration.
    use_data
        if None, then using all the gene expression profile. Else, use
        the chosen data from adata.obsm.
    copy
        Return a copy instead of writing to adata.
    Returns
    -------
    Depending on `copy`, returns or updates `adata` with the following fields.
    `X_ica` : :class:`numpy.ndarray` (`adata.obsm`)
        Independent Component Analysis representation of data.
    """

    if use_data is None:
        if issparse(adata.X):
            matrix = adata.X.toarray()
        else:
            matrix = adata.X

    else:
        matrix = adata.obsm[use_data].values

    ica = FastICA(n_components=n_factors, fun=fun, tol=tol)

    latent = ica.fit_transform(matrix)

    adata.obsm["X_ica"] = latent

    adata.uns['ica'] = {'params': {'n_factors': n_factors, 'fun': fun,
                                   'tol': tol}}

    print(
        "ICA is done! Generated in adata.obsm['X_ica'] and parameters in adata.uns['ica']")

    return adata if copy else None



from anndata import AnnData
from typing import Optional, Union
import numpy as np
import pandas as pd
from sklearn.cluster import DBSCAN
from natsort import natsorted


def localization(
    adata: AnnData,
    use_label: str = "louvain",
    eps: int = 20,
    min_samples: int = 0,
    copy: bool = False,
) -> Optional[AnnData]:

    """\
    Perform local clustering by using DBSCAN.

    Parameters
    ----------
    adata
        Annotated data matrix.
    use_label
        Use label result of clustering method.
    eps
        The maximum distance between two samples for one to be considered as 
        in the neighborhood of the other. This is not a maximum bound on the 
        distances of points within a cluster. This is the most important DBSCAN 
        parameter to choose appropriately for your data set and distance function.
    min_samples
        The number of samples (or total weight) in a neighborhood for a point to be
        considered as a core point. This includes the point itself.
    copy
        Return a copy instead of writing to adata.
    Returns
    -------
    Anndata
    """

    if "sub_cluster_labels" in adata.obs.columns:
        adata.obs = adata.obs.drop("sub_cluster_labels", axis=1)

    pd.set_option('mode.chained_assignment', None)
    subclusters = pd.Series()
    for i in adata.obs[use_label].unique():

        tmp = adata.obs[adata.obs[use_label] == i]

        clustering = DBSCAN(eps=eps, min_samples=0, algorithm="kd_tree").fit(
            tmp[["imagerow", "imagecol"]])

        labels = clustering.labels_

        sublabels = []
        for label in labels:
            sublabels.append(str(i)+"_"+str(label))
        tmp["sub_labels"] = sublabels
        subclusters = subclusters.append(tmp["sub_labels"])
    pd.reset_option('mode.chained_assignment')

    adata.obs = pd.merge(adata.obs, pd.DataFrame(
        {"sub_cluster_labels": subclusters}),
        left_index=True, right_index=True)
    # Convert to numeric
    converted = dict(enumerate(adata.obs["sub_cluster_labels"].unique()))
    inv_map = {v: k for k, v in converted.items()}
    adata.obs["sub_cluster_labels"] = adata.obs["sub_cluster_labels"].replace(
        inv_map)

    adata.obs["sub_cluster_labels"] = pd.Categorical(
        values=np.array(adata.obs["sub_cluster_labels"]).astype('U'),
        categories=natsorted(
            np.unique(np.array(adata.obs["sub_cluster_labels"])).astype('U')),
    )

    return adata if copy else None



import logging

import numpy as np
import pandas as pd
from scipy import optimize
from anndata import AnnData
import NaiveDE
from . import base


# Variational update functions

def Q_Z_expectation(mu, Y, s2e, N, C, G, pi=None):
    if pi is None:
        pi = np.ones(C) / C

    log_rho = np.log(pi[None, :]) \
              - 0.5 * N * np.log(s2e) \
              - 0.5 * np.sum((mu.T[None, :, :] - Y[:, None, :]) ** 2, 2) / s2e \
              - 0.5 * N * np.log(2 * np.pi)

    # Subtract max per row for numerical stability, and add offset from 0 for same reason.
    rho = np.exp(log_rho - log_rho.max(1)[:, None]) + 1e-12
    # Then evaluate softmax
    r = (rho.T / (rho.sum(1))).T
    
    return r


def Q_mu_k_expectation(Z_k, Y, K, s2e):
    y_k_tilde = np.dot(Z_k, Y) / s2e
    Sytk = np.dot(K, y_k_tilde)
    IpSDk = np.eye(K.shape[0]) + K * Z_k.sum() / s2e
    m_k = np.linalg.solve(IpSDk, Sytk)
    
    return m_k


def Q_mu_expectation(Z, Y, K, s2e):
    m = np.zeros((Y.shape[1], Z.shape[1]))

    y_k_tilde = np.dot(Z.T, Y).T / s2e

    for k in range(Z.shape[1]):
        m[:, k] = Q_mu_k_expectation(Z[:, k], Y, K, s2e)

    return m


# Log expectation functions for the ELBO

def ln_P_YZms(Y, Z, mu, s2e, pi=None):
    ''' Expecation of ln P(Y | Z, mu, s2e)
    '''
    G = Y.shape[0]
    N = Y.shape[1]
    C = Z.shape[1]
    if pi is None:
        pi = np.ones(C) / C
    
    log_rho = np.log(pi[None, :]) \
              - 0.5 * N * np.log(s2e) \
              - 0.5 * np.sum((mu.T[None, :, :] - Y[:, None, :]) ** 2, 2) / s2e \
              - 0.5 * N * np.log(2 * np.pi)
            
    return (Z * log_rho).sum()


def ln_P_mu(mu, K):
    ''' Expectation of ln P(mu)
    '''
    N = K.shape[0]
    C = mu.shape[1]
    ll = 0
    for k in range(C):
        ll = ll + np.linalg.det(K)
        ll = ll + mu[:, k].dot(np.linalg.solve(K, mu[:, k]))
        ll = ll + N * np.log(2 * np.pi)
        
    ll = -0.5 * ll
    
    return ll


def ln_P_Z(Z, pi=None):
    ''' Expectation of ln P(Z)
    '''
    C = Z.shape[1]
    if pi is None:
        pi = np.ones(C) / C
        
    return np.dot(Z, np.log(pi)).sum()


def ln_Q_mu(K, Z, s2e):
    ''' Expecation of ln Q(mu)
    '''
    N = K.shape[0]
    C = Z.shape[1]
    G_k = Z.sum(0)
    
    ll = 0
    U, S = base.factor(K)
    for k in range(C):
        ll = ll - (1. / S + G_k[k] / s2e).sum()
        ll = ll + N * np.log(2 * np.pi)
        
    
    ll = -0.5 * ll
    
    return ll


def ln_Q_Z(Z, r):
    ''' Expectation of ln Q(Z)
    '''
    return np.sum(Z * np.log(r))


# ELBO and ELBO objective

def ELBO(Y, r, m, s2e, K, K_0, s2e_0, pi=None):
    L = ln_P_YZms(Y, r, m, s2e, pi) + ln_P_Z(r, pi) + ln_P_mu(m, K) \
        - ln_Q_Z(r, r) - ln_Q_mu(K_0, r, s2e_0)
    
    return L


def make_elbojective(Y, r, m, X, K_0, s2e_0, pi=None):
    def elbojective(log_s2e):
        return -ELBO(Y, r, m, np.exp(log_s2e), K_0, K_0, s2e_0, pi)
    
    return elbojective


# Model fitting

def fit_patterns(X, Y, C, l, s2e_0=1.0, verbosity=0, maxiter=100, printerval=1, opt_interval=1, delta_elbo_threshold=1e-2):
    ''' Fit spatial patterns using Automatic Expression Histology.

    X : Spatial coordinates

    Y : Gene expression values

    C : The number of patterns

    l : The charancteristic length scale of the clusters

    Returns

    final_elbo : The final ELBO value.

    m : The posterior mean underlying expression values for each cluster.

    r : The posterior pattern assignment probabilities for each gene and pattern.

    s2e : The estimated noise parameter of the model
    '''
    # Set up constants
    G = Y.shape[0]
    N = Y.shape[1]
    eps = 1e-8 * np.eye(N)
    
    s2e = s2e_0
    
    K = base.SE_kernel(X, l) + eps
    
    # Randomly initialize
    r = np.random.uniform(size=(G, C))
    r = r / r.sum(0)
    
    pi = r.sum(0) / G

    m = np.random.normal(size=(N, C))
    
    elbo_0 = ELBO(Y, r, m, s2e, K, K, s2e, pi)
    elbo_1 = elbo_0

    if verbosity > 0:
        print('iter {}, ELBO: {:0.2e}'.format(0, elbo_1))

    if verbosity > 1:
        print()

    for i in range(maxiter):
        if (i % opt_interval == (opt_interval - 1)):
            elbojective = make_elbojective(Y, r, m, X, K, s2e, pi)
            
            o = optimize.minimize_scalar(elbojective)
            s2e = np.exp(o.x)
            
            
        r = Q_Z_expectation(m, Y, s2e, N, C, G, pi)
        m = Q_mu_expectation(r, Y, K, s2e)
        
        pi = r.sum(0) / G

        elbo_0 = elbo_1
        elbo_1 = ELBO(Y, r, m, s2e, K, K, s2e, pi)
        delta_elbo = np.abs(elbo_1 - elbo_0)

        if verbosity > 0 and (i % printerval == 0):
            print('iter {}, ELBO: {:0.2e}, delta_ELBO: {:0.2e}'.format(i + 1, elbo_1, delta_elbo))
            
            if verbosity > 1:
                print('ln(l): {:0.2f}, ln(s2e): {:.2f}'.format(np.log(l), np.log(s2e)))
                
            if verbosity > 2:
                line1 = 'P(Y | Z, mu, s2e): {:0.2e}, P(Z): {:0.2e}, P(mu): {:0.2e}' \
                        .format(ln_P_YZms(Y, r, m, s2e, pi), ln_P_Z(r, pi), ln_P_mu(m, K))
                line2 = 'Q(Z): {:0.2e}, Q(mu): {:0.2e}'.format(ln_Q_Z(r, r), ln_Q_mu(K, r, s2e))
                print(line1 + '\n' + line2)
            
            if verbosity > 1:
                print()
            
        if delta_elbo < delta_elbo_threshold:
            if verbosity > 0:
                print('Converged on iter {}'.format(i + 1))

            break
            
    else:
        print('Warning! ELBO dit not converge after {} iters!'.format(i + 1))

    final_elbo = ELBO(Y, r, m, s2e, K, K, s2e, pi)
        
    return final_elbo, m, r, s2e


def spatial_patterns(
    adata: AnnData, 
    results, 
    C: int = 3, 
    use_data: str = "normalized", 
    copy: bool = False,
    **kwargs):

    ''' Group spatially variable genes into spatial patterns using
    automatic expression histology (AEH).

    X : Spatial coordinates

    exp_mat : Expression matrix, appropriately normalised.

    DE_mll_results : Results table from SpatialDE, after filtering
        for significance level.

    C : The number of spatial patterns

    **kwargs are passed on to the function fit_patterns()

    Returns

    pattern_results : A DataFrame with pattern membership information
        for each gene

    patterns : The posterior mean underlying expression for genes in
        given spatial patterns.
    '''

    sample_info = pd.DataFrame(0, index = adata.obsm[use_data].index, columns = ['imagerow','imagecol','sum_counts'])
    sample_info.loc[:,'imagerow'] = adata.obs['imagerow']
    sample_info.loc[:,'imagecol'] = adata.obs['imagecol']
    sample_info.loc[:,'sum_counts'] = adata.obs['sum_counts']

    norm_expr = NaiveDE.stabilize(adata.obsm[use_data].T).T
    resid_expr = NaiveDE.regress_out(sample_info, norm_expr.T, 'np.log(sum_counts)').T

    sign_results = results.query('qval < 0.05')
    l = sign_results['l'].value_counts().argmax() * 1.05

    X = sample_info[['imagerow', 'imagecol']]
    Y = resid_expr[sign_results['g']].values.T

    # This is important, we only care about co-expression, not absolute levels.
    Y = (Y.T - Y.mean(1)).T
    Y = (Y.T / Y.std(1)).T

    _, m, r, _ = fit_patterns(X, Y, C, l, **kwargs)

    cres = pd.DataFrame({'g': sign_results['g'],
                         'pattern': r.argmax(1),
                         'membership': r.max(1)})


    adata.uns["histology_results"] = cres

    for i in range(0,C):
        adata.uns["pattern_" + str(i)] = cres.query('pattern == @i').sort_values('membership')['g'].tolist()

    return adata if copy else None



''' Wrapper functions to use SpatialDE directly on AnnData objects
'''
import logging

import pandas as pd

import NaiveDE

from .aeh import spatial_patterns
from .base import SpatialDE
from .util import qvalue

def spatialde_test(adata, coord_columns=['x', 'y'], regress_formula='np.log(total_counts)'):
    ''' Run the SpatialDE test on an AnnData object

    Parameters
    ----------

    adata: An AnnData object with counts in the .X field.

    coord_columns: A list with the columns of adata.obs which represent spatial
                   coordinates. Default ['x', 'y'].

    regress_formula: A patsy formula for linearly regressing out fixed effects
                     from columns in adata.obs before fitting the SpatialDE models.
                     Default is 'np.log(total_counts)'.

    Returns
    -------

    results: A table of spatial statistics for each gene.
    '''
    logging.info('Performing VST for NB counts')
    adata.layers['stabilized'] = NaiveDE.stabilize(adata.X.T).T

    logging.info('Regressing out fixed effects')
    adata.layers['residual'] = NaiveDE.regress_out(adata.obs,
                                                   adata.layers['stabilized'].T,
                                                   regress_formula).T

    X = adata.obs[coord_columns].values
    expr_mat = pd.DataFrame.from_records(adata.layers['residual'],
                                         columns=adata.var.index,
                                         index=adata.obs.index)

    results = SpatialDE(X, expr_mat)

    # Clip 0 pvalues
    min_pval = results.query('pval > 0')['pval'].min() / 2
    results['pval'] = results['pval'].clip_lower(min_pval)

    # Correct for multiple testing
    results['qval'] = qvalue(results['pval'], pi0=1.)

    return results


def automatic_expression_histology(adata, filtered_results, C, l,
                                    coord_columns=['x', 'y'], layer='residual', **kwargs):
    ''' Fit the Automatic Expression Histology (AEH) model to
    expression in an AnnData object.

    Parameters
    ----------

    adata: An AnnData object with a layer of stabilized expression values

    filtered_results: A DataFrame with the signifificant subset of results
                      from the SpatialDE significance test.

    C: integer, the number of hidden spatial patterns.

    l: float, the common lengthscale for the hidden spatial patterns.

    coord_columns: A list with the columns of adata.obs which represent spatial
                   coordinates. Default ['x', 'y'].

    layer: A string indicating the layer of adata to fit the AEH model to.
           By defualt uses the 'residual' layer.

    Remaining arguments are passed to SpatialDE.aeh.spatial_patterns()

    Returns
    -------

    (histology_results, patterns)

    histology_results: DataFrame with pattern membership information for each gene.

    patterns: DataFrame with the inferred hidden spatial functions the genes belong to
              evaluated at all points in the data.

    '''
    X = adata.obs[coord_columns].values

    expr_mat = pd.DataFrame.from_records(adata.layers[layer],
                                         columns=adata.var.index,
                                         index=adata.obs.index)

    logging.info('Performing Automatic Expression Histology')
    histology_results, patterns = spatial_patterns(X, expr_mat, filtered_results,
                                                   C, l, **kwargs)

    return histology_results, patterns


from anndata import AnnData
import stlearn
import os

def SMEclust(
    adata: AnnData,
    out_path: str = "../tiling",
    crop_size: float = 40,
    use_data: str = "X_pca",
    radius: float = 50,
    adjust_method: str = "mean",
    n_neighbors: int = 25,
    use_label: str = "louvain",
    n_clusters: int = 10,
    random_state: int = 0,
    copy: bool = False,

    ) -> AnnData:

	if not os.path.isdir(out_path):
		os.mkdir(out_path)

	stlearn.pp.tiling(adata,out_path=out_path,crop_size = crop_size)
	stlearn.pp.extract_feature(adata)
	stlearn.spatial.morphology.adjust(adata,
		use_data=use_data,
		radius=radius,
		method=adjust_method)

	stlearn.pp.neighbors(adata,
		n_neighbors=n_neighbors,
		use_rep=use_data + '_morphology',
		random_state=random_state)

	if use_label == "louvain":
		stlearn.tl.clustering.louvain(adata,
			random_state=random_state)
	elif use_label == "kmeans":
		stlearn.tl.clustering.kmeans(adata,
			use_data = use_data,
			n_clusters=n_clusters)
	else:
		raise ValueError("Wrong method!")

	return adata if copy else None



from anndata import AnnData
from typing import Optional, Union
import numpy as np
import pandas as pd
import networkx as nx
from scipy.spatial.distance import cdist


def global_level(
    adata: AnnData,
    use_label: str = "louvain",
    list_cluster: list = [],
    w: float = 0.5,

    copy: bool = False,
) -> Optional[AnnData]:

    """\
    Perform global sptial trajectory inference.

    Parameters
    ----------
    adata
        Annotated data matrix.
    list_cluster
        Setup a list of cluster to perform pseudo-space-time
    use_label
        Use label result of clustering method.
    w
        Pseudo-spatio-temporal distance weight (balance between spatial effect and DPT)
    copy
        Return a copy instead of writing to adata.
    Returns
    -------
    Anndata
    """

    assert w <= 1, "w should be in range 0 to 1"
    # Get global graph
    G = adata.uns["global_graph"]
    # Convert to directed graph
    H = G.to_directed()

    # Query cluster
    query_nodes = list_cluster
    query_nodes = ordering_nodes(query_nodes, adata)

    query_dict = {}
    order_dict = {}

    for i in query_nodes:
        order = 0
        for j in adata.obs[adata.obs[use_label] == str(i)]["sub_cluster_labels"].unique():
            query_dict[int(j)] = int(i)
            order_dict[int(j)] = int(order)

            order += 1

    dm_list = []
    sdm_list = []
    order_big_dict = {}
    edge_list = []
    for i, j in enumerate(query_nodes):
        order_big_dict[j] = int(i)
        if i == len(query_nodes)-1:
            break
        for j in adata.uns["split_node"][query_nodes[i]]:
            for k in adata.uns["split_node"][query_nodes[i+1]]:
                edge_list.append((int(j), int(k)))

        # Calculate DPT distance matrix
        dm_list.append(dpt_distance_matrix(
            adata, query_nodes[i], query_nodes[i+1], use_label=use_label))
        # Calculate Spatial distance matrix
        sdm_list.append(spatial_distance_matrix(
            adata, query_nodes[i], query_nodes[i+1], use_label=use_label))

    # Get centroid dictionary
    centroid_dict = adata.uns["centroid_dict"]

    H_sub = H.edge_subgraph(edge_list)
    H_sub = nx.DiGraph(H_sub)
    prepare_root = []
    for node in adata.uns["split_node"][query_nodes[0]]:
        H_sub.add_edge(9999, int(node))
        prepare_root.append(centroid_dict[int(node)])

    prepare_root = np.array(prepare_root)
    centroide = (sum(prepare_root[:, 0])/len(prepare_root[:, 0]),
                 sum(prepare_root[:, 1])/len(prepare_root[:, 1]))
    centroid_dict[9999] = centroide

    labels = nx.get_edge_attributes(H_sub, 'weight')

    for edge, _ in labels.items():

        dm = dm_list[order_big_dict[query_dict[edge[0]]]]
        sdm = sdm_list[order_big_dict[query_dict[edge[0]]]]

        weight = dm[order_dict[edge[0]], order_dict[edge[1]]] * \
            w + sdm[order_dict[edge[0]], order_dict[edge[1]]]*(1-w)
        H_sub[edge[0]][edge[1]]['weight'] = weight

    H_sub = nx.algorithms.tree.minimum_spanning_arborescence(H_sub)
    #remove = [edge for edge in H_sub.edges if 9999 in edge]
    # H_sub.remove_edges_from(remove)
    # remove.remove_node(9999)

    adata.uns["PTS_graph"] = H_sub


def get_node(node_list, split_node):
    result = np.array([])
    for node in node_list:
        result = np.append(result, np.array(split_node[node]).astype(int))
    return result.astype(int)


def ordering_nodes(node_list, adata):
    mean_dpt = []
    for node in node_list:
        mean_dpt.append(adata.obs[adata.obs["louvain"]
                                  == str(node)]["dpt_pseudotime"].mean())

    return list(np.array(node_list)[np.argsort(mean_dpt)])


def dpt_distance_matrix(adata, cluster1, cluster2, use_label):
    tmp = adata.obs[adata.obs[use_label] == str(cluster1)]
    chosen_adata1 = adata[list(tmp.index)]
    tmp = adata.obs[adata.obs[use_label] == str(cluster2)]
    chosen_adata2 = adata[list(tmp.index)]

    sub_dpt1 = []
    chosen_sub1 = chosen_adata1.obs["sub_cluster_labels"].unique()
    for i in range(0, len(chosen_sub1)):
        sub_dpt1.append(
            chosen_adata1.obs[chosen_adata1.obs["sub_cluster_labels"] == chosen_sub1[i]]["dpt_pseudotime"].max())

    sub_dpt2 = []
    chosen_sub2 = chosen_adata2.obs["sub_cluster_labels"].unique()
    for i in range(0, len(chosen_sub2)):
        sub_dpt2.append(
            chosen_adata2.obs[chosen_adata2.obs["sub_cluster_labels"] == chosen_sub2[i]]["dpt_pseudotime"].max())

    dm = cdist(np.array(sub_dpt1).reshape(-1, 1),
               np.array(sub_dpt2).reshape(-1, 1), lambda u, v: np.abs(u-v))
    #from sklearn.preprocessing import MinMaxScaler
    #scaler = MinMaxScaler()
    #scale_dm = scaler.fit_transform(dm)
    scale_dm = dm/np.max(dm)
    return scale_dm


def spatial_distance_matrix(adata, cluster1, cluster2, use_label):

    tmp = adata.obs[adata.obs[use_label] == str(cluster1)]
    chosen_adata1 = adata[list(tmp.index)]
    tmp = adata.obs[adata.obs[use_label] == str(cluster2)]
    chosen_adata2 = adata[list(tmp.index)]

    centroid_dict = adata.uns["centroid_dict"]

    sub_coord1 = []
    chosen_sub1 = chosen_adata1.obs["sub_cluster_labels"].unique()
    for i in range(0, len(chosen_sub1)):
        sub_coord1.append(centroid_dict[int(chosen_sub1[i])])

    sub_coord2 = []
    chosen_sub2 = chosen_adata2.obs["sub_cluster_labels"].unique()
    for i in range(0, len(chosen_sub2)):
        sub_coord2.append(centroid_dict[int(chosen_sub2[i])])

    sdm = cdist(np.array(sub_coord1), np.array(sub_coord2), "euclidean")

    #from sklearn.preprocessing import MinMaxScaler
    #scaler = MinMaxScaler()
    #scale_sdm = scaler.fit_transform(sdm)
    scale_sdm = sdm/np.max(sdm)

    return scale_sdm



import logging

import numpy as np
import pandas as pd

from . import base


def optimize_gp_params(Ymean, Yvar, Us, UT1s, Ss, Ks, N):
    ''' Optimize GP parameters for pattern of (Ymean, Yvar)
        over the covariance structure grid.
    '''
    ll_max = -np.inf
    top_k = 0
    par_max = [-np.inf, 1., 1.]
    k = 0
    for U, UT1, S, K in zip(Us, UT1s, Ss, Ks):
        UTy = base.get_UTy(U, Ymean[:, 0])
        ll, delta, mu, s2s, _ = base.lbfgsb_max_LL(UTy, UT1, S, N, Yvar)

        if ll > ll_max:
            ll_max = ll
            top_k = k
            par_max = [ll, delta, s2s]

        k += 1

    ll, delta, s2s = par_max
    
    return ll, delta, s2s, top_k


def gp_smoothing(Ymean, Yvar, K, s2s, delta):
    noise_K = s2s * K + delta * s2s * np.diag(Yvar)
    Y_hat = s2s * K.dot(np.linalg.solve(noise_K, Ymean))
    return Y_hat


def aeh(X, Y, C, Ks=None, max_iter=10):
    # We're only interested in patterns, so center data
    Y = (Y.T - Y.mean(1)).T

    # Pre-factor kernel search space
    Us, Ss, UT1s = [], [], []
    for K in Ks:
        U, S = base.factor(K)
        Us.append(U)
        Ss.append(S)
        UT1 = base.get_UT1(U)
        UT1s.append(UT1)

    N = K.shape[0]
    G = Y.shape[0]

    # Initialize by random assignment
    idx = np.arange(G)
    np.random.shuffle(idx)
    cidxs = np.array_split(idx, C)
    new_clusts = np.zeros(len(idx))

    for j in range(max_iter):
        finish = False
        Yhats = []
        ll_sum = 0
        params = []
        for genes in cidxs:
            if len(genes) == 0:
                continue

            # Make average pattern per cluster, with observation variance
            Ymean = Y[genes].mean(0)[:, None]
            Yvar = Y[genes].var(0) + 1.

            # Learn noise level and lengthscale of pattern
            ll, delta, s2s, top_k = optimize_gp_params(Ymean, Yvar, Us, UT1s, Ss, Ks, N)
            K = Ks[top_k]


            # Create predicted average pattern
            Yhat = gp_smoothing(Ymean, Yvar, K, s2s, delta)
            Yhats.append(Yhat)

        # Calculate residuals for each pattern and gene
        tYhats = np.dstack(Yhats)
        tY = np.transpose(Y[:, :, None], (1, 0, 2))
        cost = np.square(tY - tYhats).sum(0)

        # Switch clusters for genes to minimize residuals
        if not finish:
            old_clusts = new_clusts.copy()
            new_clusts = cost.argmin(1)

            if np.array_equal(old_clusts, new_clusts):
                logging.info('Converged! Finishing...')
                finish = True

            cidxs = [np.where(new_clusts == i)[0] for i in range(C)]

        # Calculate the new data likelihood
        log_likelihood = -(N + G) / 2. * np.log(2. * np.pi) - \
                          (N + G) / 2. * np.log(cost.min(1).sum() / (N + G)) - \
                          (N + G) / 2.

        logging.info('Iteration {}, log likelihood: {:.2f}'.format(j + 1, log_likelihood))

        if finish:
            break

    return new_clusts, cost, Yhats, log_likelihood


def spatial_patterns(X, exp_mat, DE_mll_results, C, max_iter=10, kernel_space=None):
    ''' Group spatially variable genes into spatial patterns using automatic
     histology analysis (AHA).

    Returns
    
    pattern_results : A DataFrame with pattern membership information
        for each gene

    patterns : The spatial patterns underlying the expression values
        for the genes in the given pattern.

    '''
    if kernel_space == None:
        l_min, l_max = base.get_l_limits(X)
        l_range = np.logspace(np.log10(l_min), np.log10(l_max), 10)

        Ks = [base.SE_kernel(X, l) for l in l_range]
        
    else:
        raise NotImplementedError('Custom kernels not supported for AEH.')
        
    Y = exp_mat[DE_mll_results['g']].values.T
    N = Y.shape[1]
    
    patterns, cost, Yhats, ll = aeh(X, Y, C, Ks=Ks, max_iter=max_iter)
    
    cres = pd.DataFrame({'g': DE_mll_results['g'],
                         'pattern': patterns,
                         'membership': cost.min(1) / N})
    
    Yhats = np.hstack(Yhats)
    Yhats = pd.DataFrame.from_records(Yhats)
    Yhats.index = exp_mat.index
    
    return cres, Yhats



from typing import Optional, Union
from anndata import AnnData
from matplotlib import pyplot as plt
from pathlib import Path
import os
import sys
import numpy as np

def parsing(
    adata: AnnData,
    coordinates_file: Union[Path, str],
    copy: bool = True,
) -> Optional[AnnData]:
    
    """\
    Parsing the old spaital transcriptomics data

    Parameters
    ----------
    adata
        Annotated data matrix.
    coordinates_file
        Coordinate file generated by st_spot_detector.
    copy
        Return a copy instead of writing to adata.
    Returns
    -------
    Depending on `copy`, returns or updates `adata` with the following fields.
    **imagecol** and **imagerow** : `adata.obs` field
        Spatial information of the tissue image.
    """

    # Get a map of the new coordinates
    new_coordinates = dict()
    with open(coordinates_file, "r") as filehandler:
        for line in filehandler.readlines():
            tokens = line.split()
            assert(len(tokens) == 6 or len(tokens) == 4)
            if tokens[0] != "x":
                old_x = int(tokens[0])
                old_y = int(tokens[1])
                new_x = round(float(tokens[2]), 2)
                new_y = round(float(tokens[3]), 2)
                if len(tokens) == 6:
                    pixel_x = float(tokens[4])
                    pixel_y = float(tokens[5])
                    new_coordinates[(old_x, old_y)] = (pixel_x, pixel_y)
                else:
                    raise ValueError("Error, output format is pixel coordinates but\n "
                                     "the coordinates file only contains 4 columns\n")

    counts_table = adata.to_df()
    new_index_values = list()

    imgcol = []
    imgrow = []
    for index in counts_table.index:
        tokens = index.split("x")
        x = int(tokens[0])
        y = int(tokens[1])
        try:
            new_x, new_y = new_coordinates[(x, y)]
            imgcol.append(new_x)
            imgrow.append(new_y)

            new_index_values.append("{0}x{1}".format(new_x, new_y))
        except KeyError:
            counts_table.drop(index, inplace=True)

    # Assign the new indexes
    #counts_table.index = new_index_values

    # Remove genes that have now a total count of zero
    counts_table = counts_table.transpose(
    )[counts_table.sum(axis=0) > 0].transpose()

    adata = AnnData(counts_table)

    adata.obs["imagecol"] = imgcol
    adata.obs["imagerow"] = imgrow

    adata.obsm["spatial"] = np.c_[[imgcol,imgrow]].reshape(-1,2)

    return adata if copy else None



from anndata import AnnData
from typing import Optional, Union
import numpy as np
from stlearn.em import run_pca, run_diffmap
from stlearn.pp import neighbors
from scipy.spatial.distance import cdist


def local_level(
    adata: AnnData,
    use_label: str = "louvain",
    cluster: int = 9,
    w: float = 0.5,
    copy: bool = False,
) -> Optional[AnnData]:

    """\
    Perform local sptial trajectory inference (required run pseudotime first).

    Parameters
    ----------
    adata
        Annotated data matrix.
    use_label
        Use label result of clustering method.
    cluster
        Choose cluster to perform local spatial trajectory inference.
    threshold
        Threshold to find the significant connection for PAGA graph.
    w
        Pseudo-spatio-temporal distance weight (balance between spatial effect and DPT)
    copy
        Return a copy instead of writing to adata.
    Returns
    -------
    Anndata
    """

    print("Start construct trajectory for subcluster " + str(cluster))

    tmp = adata.obs[adata.obs[use_label] == str(cluster)]
    cluster_data = adata[list(tmp.index)]

    list_cluster = cluster_data.obs["sub_cluster_labels"].unique()
    dpt = []
    sd = []
    centroid_dict = cluster_data.uns["centroid_dict"]
    for i in list_cluster:
        dpt.append(
            cluster_data.obs[cluster_data.obs["sub_cluster_labels"] == i]["dpt_pseudotime"].max())
        sd.append(centroid_dict[int(i)])
    dm = cdist(np.array(dpt).reshape(-1, 1),
               np.array(dpt).reshape(-1, 1), lambda u, v: np.abs(u-v))

    non_abs_dm = cdist(np.array(dpt).reshape(-1, 1),
                       np.array(dpt).reshape(-1, 1), lambda u, v: u-v)
    adata.uns["nonabs_dpt_distance_matrix"] = non_abs_dm

    scale_dm = dm/np.max(dm)
    sdm = cdist(np.array(sd), np.array(sd), "euclidean")
    scale_sdm = sdm/np.max(sdm)

    stdm = scale_dm*w + scale_sdm*(1-w)
    adata.uns["ST_distance_matrix"] = stdm

    return adata if copy else None



from types import MappingProxyType
from typing import Optional, Tuple, Sequence, Type, Mapping, Any, Union

import numpy as np
import pandas as pd
from anndata import AnnData
from natsort import natsorted
from numpy.random.mtrand import RandomState
from scipy.sparse import spmatrix
from stlearn._compat import Literal

try:
    from louvain.VertexPartition import MutableVertexPartition
except ImportError:
    class MutableVertexPartition: pass
    MutableVertexPartition.__module__ = 'louvain.VertexPartition'
import scanpy

def louvain(
    adata: AnnData,
    resolution: Optional[float] = None,
    random_state: Optional[Union[int, RandomState]] = 0,
    restrict_to: Optional[Tuple[str, Sequence[str]]] = None,
    key_added: str = 'louvain',
    adjacency: Optional[spmatrix] = None,
    flavor: Literal['vtraag', 'igraph', 'rapids'] = 'vtraag',
    directed: bool = True,
    use_weights: bool = False,
    partition_type: Optional[Type[MutableVertexPartition]] = None,
    partition_kwargs: Mapping[str, Any] = MappingProxyType({}),
    copy: bool = False,
) -> Optional[AnnData]:
    """\
    Wrap function scanpy.tl.louvain
    Cluster cells into subgroups [Blondel08]_ [Levine15]_ [Traag17]_.
    Cluster cells using the Louvain algorithm [Blondel08]_ in the implementation
    of [Traag17]_. The Louvain algorithm has been proposed for single-cell
    analysis by [Levine15]_.
    This requires having ran :func:`~scanpy.pp.neighbors` or
    :func:`~scanpy.external.pp.bbknn` first,
    or explicitly passing a ``adjacency`` matrix.
    Parameters
    ----------
    adata
        The annotated data matrix.
    resolution
        For the default flavor (``'vtraag'``), you can provide a resolution
        (higher resolution means finding more and smaller clusters),
        which defaults to 1.0.
        See “Time as a resolution parameter” in [Lambiotte09]_.
    random_state
        Change the initialization of the optimization.
    restrict_to
        Restrict the clustering to the categories within the key for sample
        annotation, tuple needs to contain ``(obs_key, list_of_categories)``.
    key_added
        Key under which to add the cluster labels. (default: ``'louvain'``)
    adjacency
        Sparse adjacency matrix of the graph, defaults to
        ``adata.uns['neighbors']['connectivities']``.
    flavor
        Choose between to packages for computing the clustering.
        ``'vtraag'`` is much more powerful, and the default.
    directed
        Interpret the ``adjacency`` matrix as directed graph?
    use_weights
        Use weights from knn graph.
    partition_type
        Type of partition to use.
        Only a valid argument if ``flavor`` is ``'vtraag'``.
    partition_kwargs
        Key word arguments to pass to partitioning,
        if ``vtraag`` method is being used.
    copy
        Copy adata or modify it inplace.
    Returns
    -------
    :obj:`None`
        By default (``copy=False``), updates ``adata`` with the following fields:
        ``adata.obs['louvain']`` (:class:`pandas.Series`, dtype ``category``)
            Array of dim (number of samples) that stores the subgroup id
            (``'0'``, ``'1'``, ...) for each cell.
    :class:`~anndata.AnnData`
        When ``copy=True`` is set, a copy of ``adata`` with those fields is returned.
    """

    scanpy.tl.louvain(adata, resolution=resolution, random_state=random_state, 
            restrict_to=restrict_to, key_added=key_added, adjacency=adjacency,
            flavor=flavor, directed=directed, use_weights=use_weights, partition_type=partition_type,
            partition_kwargs=partition_kwargs, copy=copy)

    print("Applying Louvain clustering ...")
    print("Louvain clustering is done! The labels are stored in adata.obs['louvain']")


import numpy as np
import pandas as pd
import scipy as sc
import scipy.spatial as spatial
from anndata import AnnData
from ...clustering.louvain import louvain
from ....preprocessing.graph import neighbors


# cluster spatial spots based on the proportion of known ligand-receptor co-expression among the neighbouring spots
def lr(
    adata: AnnData,
    use_data: str,
    threshold: float = 0,
    distance: int = 10,
) -> AnnData:
    """ cluster spatial spots based on the proportion of known ligand-receptor co-expression among the neighbouring spots
    Parameters
    ----------
    adata: AnnData          The data object to scan
    use_data: str           Data to be used in L-R scanning
    threshold: float        Threshold to determine the significant L-R expression in counting
    distance: int           Distance to determine the nearest neighbour
    
    Returns
    -------
    adata: AnnData          The data object including the lr_scan results
    """
    data = adata.obsm[use_data]
    if not isinstance(data, pd.DataFrame):
        if sc.sparse.issparse(data):
            data = pd.DataFrame(data.toarray(), index=adata.obs_names, columns=adata.var_names)
        else:
            data = pd.DataFrame(data, index=adata.obs_names, columns=adata.var_names)

    lr_pairs = adata.uns['lr'].copy()
    lr_pairs += [item.split('_')[1]+'_'+item.split('_')[0] for item in lr_pairs]

    # get neighbour spots for each spot
    coor = adata.obs[['imagerow', 'imagecol']]
    point_tree = spatial.cKDTree(coor)
    neighbours = []
    for spot in adata.obs_names:
        n_index = point_tree.query_ball_point(np.array([adata.obs['imagerow'].loc[spot], adata.obs['imagecol'].loc[spot]]), distance)
        neighbours.append([item for item in data.index[n_index] if item != spot])
    
    # filter out those LR not existing in the dataset
    ligands = [item.split('_')[0] for item in lr_pairs]
    receptors = [item.split('_')[1] for item in lr_pairs]
    avail = [i for i, x in enumerate(ligands) if ligands[i] in data.columns and receptors[i] in data.columns]   
    spot_ligands = data.loc[:, [ligands[i] for i in avail]]
    spot_receptors = data.loc[:, [receptors[i] for i in avail]]
    print('Altogether ' + str(len(avail)) + ' valid L-R pairs')

    # count the co-expressed ligand-recptor pairs between neighbours
    def count_receptors(x):
        nbs = spot_receptors.loc[neighbours[data.index.tolist().index(x.name)], :]
        return (nbs > threshold).sum(axis=0) / nbs.shape[0]

    def count_ligands(x):
        nbs = spot_ligands.loc[neighbours[data.index.tolist().index(x.name)], :]
        return (nbs > threshold).sum(axis=0) / nbs.shape[0]

    nb_receptors = spot_receptors.apply(count_receptors, axis=1)   # proportion of neighbour spots which has receptor expression > threshold
    nb_ligands = spot_ligands.apply(count_ligands, axis=1)   # proportion of neighbour spots which has receptor expression > threshold
    # ligands on the spots
    st_lr_neighbour_ligands = pd.DataFrame((spot_ligands > threshold).values * nb_receptors.values, index=data.index, columns=[lr_pairs[i] for i in avail])
    # receptors on the spots
    st_lr_neighbour_receptors = pd.DataFrame((spot_receptors > threshold).values * nb_ligands.values, index=data.index, columns=[lr_pairs[i] for i in avail])
    adata.obsm['lr_neighbours'] = st_lr_neighbour_ligands + st_lr_neighbour_receptors
    print('L-R interactions with neighbours are counted and stored into adata[\'lr_neighbours\']')

    neighbors(adata,n_neighbors=25,use_rep='lr_neighbours')
    louvain(adata, key_added='lr_neighbours_louvain')
    
    # locate the highest Ligand-Receptor expressing cluster
    st_lr_cluster = []
    for n in range(max([int(i) for i in adata.obs['lr_neighbours_louvain']]) + 1):
        spot_idx = [i for i in range(len(adata.obs['lr_neighbours_louvain'])) if int(adata.obs['lr_neighbours_louvain'][i])==n]
        st_lr_cluster.append(adata.obsm['lr_neighbours'].iloc[spot_idx, :].sum().sum() / len(spot_idx))

    adata.uns['lr_neighbours_louvain_max'] = str(st_lr_cluster.index(max(st_lr_cluster)))
    print("Spatial distribution of LR co-expression is written to adata.obsm['lr_neighbours']")
    print("Result of LR-clustering is kept in adata.obs['lr_neighbours_louvain']")
    print("The largest expressed LR neighbouring cluster is: ", adata.uns['lr_neighbours_louvain_max'])

    return adata



"""A sample module."""


def feet_to_meters(feet):
    """Convert feet to meters."""
    try:
        value = float(feet)
    except ValueError:
        print("Unable to convert to float: %s", feet)
    else:
        return (0.3048 * value * 10000.0 + 0.5) / 10000.0



from typing import Union, Optional, Tuple, Collection, Sequence, Iterable
from anndata import AnnData
import numpy as np
from scipy.sparse import issparse, isspmatrix_csr, csr_matrix, spmatrix
from scipy import sparse
from stlearn import logging as logg
import scanpy

def log1p(
    adata: Union[AnnData, np.ndarray, spmatrix],
    copy: bool = False,
    chunked: bool = False,
    chunk_size: Optional[int] = None,
    base: Optional[float] = None,
) -> Optional[AnnData]:
    """\
    Wrap function of scanpy.pp.log1p
    Copyright (c) 2017 F. Alexander Wolf, P. Angerer, Theis Lab

    Logarithmize the data matrix.
    Computes :math:`X = \\log(X + 1)`,
    where :math:`log` denotes the natural logarithm unless a different base is given.
    Parameters
    ----------
    data
        The (annotated) data matrix of shape `n_obs` × `n_vars`.
        Rows correspond to cells and columns to genes.
    copy
        If an :class:`~anndata.AnnData` is passed, determines whether a copy
        is returned.
    chunked
        Process the data matrix in chunks, which will save memory.
        Applies only to :class:`~anndata.AnnData`.
    chunk_size
        `n_obs` of the chunks to process the data in.
    base
        Base of the logarithm. Natural logarithm is used by default.
    Returns
    -------
    Returns or updates `data`, depending on `copy`.
    """


    scanpy.pp.log1p(adata, copy=copy, chunked=chunked, chunk_size=chunk_size, base=base)

    print("Log transformation step is finished in adata.X")


def scale(
    adata: Union[AnnData, np.ndarray, spmatrix],
    zero_center: bool = True,
    max_value: Optional[float] = None,
    copy: bool = False,
) -> Optional[AnnData]:
    """\
    Wrap function of scanpy.pp.scale

    Scale data to unit variance and zero mean.
    .. note::
        Variables (genes) that do not display any variation (are constant across
        all observations) are retained and set to 0 during this operation. In
        the future, they might be set to NaNs.
    Parameters
    ----------
    data
        The (annotated) data matrix of shape `n_obs` × `n_vars`.
        Rows correspond to cells and columns to genes.
    zero_center
        If `False`, omit zero-centering variables, which allows to handle sparse
        input efficiently.
    max_value
        Clip (truncate) to this value after scaling. If `None`, do not clip.
    copy
        If an :class:`~anndata.AnnData` is passed,
        determines whether a copy is returned.
    Returns
    -------
    Depending on `copy` returns or updates `adata` with a scaled `adata.X`.
    """


    scanpy.pp.scale(adata, zero_center=zero_center, max_value=max_value, copy=copy)

    print("Scale step is finished in adata.X")



from matplotlib import pyplot as plt
from PIL import Image
import pandas as pd
import matplotlib
import numpy as np

from stlearn._compat import Literal
from typing import Optional, Union
from anndata import AnnData
import warnings
from .utils import get_img_from_fig


def microenv_plot(
    adata: AnnData,
    use_data: str = None,
    library_id: str = None,
    data_alpha: float = 1.0,
    tissue_alpha: float = 1.0,
    cmap: str = "Spectral_r",
    spot_size: Union[float, int] = 6.5,
    show_color_bar: bool = True,
    show_axis: bool = False,
    dpi: int = 192,
    name: str = None,
    output: str = None,
    copy: bool = False,
) -> Optional[AnnData]:
    """\
    Plotting microenvironment.

    Parameters
    ----------
    adata
        Annotated data matrix.
    use_data
        Use dimensionality reduction result data.
    library_id
        Library id stored in AnnData.
    data_alpha
        Opacity of the spot.
    tissue_alpha
        Opacity of the tissue.
    cmap
        Color map to use.
    spot_size
        Size of the spot.
    show_color_bar
        Show color bar or not.
    show_axis
        Show axis or not.
    show_legend
        Show legend or not.
    dpi
        Set dpi as the resolution for the plot.
    name
        Name of the output figure file.
    output
        Save the figure as file or not.
    copy
        Return a copy instead of writing to adata.
    Returns
    -------
    Nothing
    """

    colors = _microenv_plot(adata, use_data)

    n_factor = len(colors)
    plt.ioff()

    if "plots" not in adata.uns:
        adata.uns['plots'] = {}

    adata.uns['plots'].update({use_data: {}})

    for i in range(0, n_factor):
        fig, a = plt.subplots()
        vmin = min(colors[i])
        vmax = max(colors[i])
        sc = a.scatter(adata.obs["imagecol"], adata.obs["imagerow"], edgecolor="none", alpha=data_alpha, s=spot_size, marker="o",
                       vmin=vmin, vmax=vmax, cmap=plt.get_cmap(cmap), c=colors[i])

        if show_color_bar:
            cb = plt.colorbar(sc, cax=fig.add_axes([0.78, 0.3, 0.03, 0.38]))
            cb.outline.set_visible(False)
        if not show_axis:
            a.axis('off')

        if library_id is None:
            library_id = list(adata.uns["spatial"].keys())[0]

        image = adata.uns["spatial"][library_id]["images"][adata.uns["spatial"]["use_quality"]]
        # Overlay the tissue image
        a.imshow(image, alpha=tissue_alpha, zorder=-1,)

        if name is None:
            name = method
        if output is not None:
            fig.savefig(output + "/factor_" + str(i+1) + ".png",
                        dpi=dpi, bbox_inches='tight', pad_inches=0)

        fig_np = get_img_from_fig(fig, dpi)

        plt.close(fig)

        current_plot = {"factor_"+str(i+1): fig_np}

        adata.uns['plots'][use_data].update(current_plot)

    print("The plot stored in adata.uns['plots']['" + use_data + "']")


def _microenv_plot(adata, use_data):

    n_factor = adata.obsm[use_data].shape[1]
    l_colors = []
    for i in range(0, n_factor):
        colors = adata.obsm[use_data][:, i]
        vmin = min(colors)
        vmax = max(colors)
        l_colors.append(colors)

    return l_colors



from matplotlib import pyplot as plt
import numpy as np
from typing import Optional, Union
from anndata import AnnData


def QC_plot(
        adata: AnnData,
        library_id: str = None,
        name: str = None,
        data_alpha: float = 0.8,
        tissue_alpha: float = 1.0,
        cmap: str = "Spectral_r",
        spot_size: tuple = (5, 40),
        show_color_bar: bool = True,
        show_size_legend: bool = True,
        show_axis: bool = False,
        dpi: int = 192,
        output: str = None,
) -> Optional[AnnData]:
    """\
        QC plot for sptial transcriptomics data.

        Parameters
        ----------
        adata
            Annotated data matrix.
        library_id
            Library id stored in AnnData.
        data_alpha
            Opacity of the spot.
        tissue_alpha
            Opacity of the tissue.
        cmap
            Color map to use.
        spot_size
            Size of the spot (min, max).
        show_color_bar
            Show color bar or not.
        show_axis
            Show axis or not.
        show_size_legend
            Show size legend or not.
        dpi
            Set dpi as the resolution for the plot.
        name
            Name of the output figure file.
        output
            Save the figure as file or not.
        copy
            Return a copy instead of writing to adata.
        Returns
        -------
        Nothing
        """
    from sklearn.preprocessing import MinMaxScaler
    reads_per_spot = adata.to_df().sum(axis=1)
    scaler = MinMaxScaler(feature_range=spot_size)
    reads_per_spot_size = scaler.fit_transform(
        reads_per_spot.to_numpy().reshape(-1, 1))
    genes_per_spot = adata.to_df().astype(bool).sum(axis=1)

    plt.rcParams['figure.dpi'] = dpi

    # Option for turning off showing figure
    plt.ioff()

    # Initialize matplotlib
    fig, a = plt.subplots()

    vmin = min(genes_per_spot)
    vmax = max(genes_per_spot)
    # Plot scatter plot based on pixel of spots
    plot = a.scatter(adata.obs["imagecol"], adata.obs["imagerow"], edgecolor="none", alpha=data_alpha,
                     s=reads_per_spot_size, marker="o",
                     vmin=vmin, vmax=vmax, cmap=plt.get_cmap(cmap), c=genes_per_spot)

    if show_color_bar:
        cb = plt.colorbar(plot, cax=fig.add_axes(
            [0.85, 0.5, 0.03, 0.35]), cmap=cmap)
        cb.ax.set_xlabel('Number of Genes', fontsize=10)
        cb.ax.xaxis.set_label_coords(0.98, 1.10)
        cb.outline.set_visible(False)

    if show_size_legend:
        size_min, size_max = spot_size
        markers = [size_min, size_min + 1 / 3 * (size_max - size_min),
                   size_min + 2 / 3 * (size_max - size_min), size_max]
        legend_markers = [plt.scatter([], [], s=i, c="grey") for i in markers]
        labels = [str(int(scaler.inverse_transform(np.array(i).reshape(1, 1))))
                  for i in markers]
        fig.legend(handles=legend_markers, labels=labels, loc='lower right', bbox_to_anchor=(0.75, 0.1),
                   scatterpoints=1, frameon=False, handletextpad=0.1, title="Number of Reads")

    if not show_axis:
        a.axis('off')
    if library_id is None:
        library_id = list(adata.uns["spatial"].keys())[0]

    image = adata.uns["spatial"][library_id]["images"][adata.uns["spatial"]["use_quality"]]
    # Overlay the tissue image
    a.imshow(image, alpha=tissue_alpha, zorder=-1, )

    # fig.tight_layout()
    if output is not None:
        fig.savefig(output + "/" + name + ".png", dpi=dpi,
                    bbox_inches='tight', pad_inches=0)

    plt.show()



from .preprocessing.filter_genes import filter_genes
from .preprocessing.normalize import normalize_total
from .preprocessing.log_scale import log1p
from .preprocessing.log_scale import scale
from .preprocessing.graph import neighbors
from .image_preprocessing.image_tiling import tiling
from .image_preprocessing.feature_extractor import extract_feature


from matplotlib import pyplot as plt
from PIL import Image
import pandas as pd
import matplotlib
import numpy as np

from stlearn._compat import Literal
from typing import Optional, Union
from anndata import AnnData
import warnings


def subcluster_plot(
    adata: AnnData,
    library_id: str = None,
    use_label: str = "louvain",
    cluster: int = 0,
    data_alpha: float = 1.0,
    tissue_alpha: float = 1.0,
    cmap: str = "jet",
    spot_size: Union[float, int] = 5,
    show_axis: bool = False,
    dpi: int = 192,
    name: str = None,
    output: str = None,
    copy: bool = False,
) -> Optional[AnnData]:
    """\
    Sub-clustering plot for sptial transcriptomics data.

    Parameters
    ----------
    adata
        Annotated data matrix.
    library_id
        Library id stored in AnnData.
    use_label
        Use label result of clustering method.
    list_cluster
        Choose set of clusters that will display in the plot.
    data_alpha
        Opacity of the spot.
    tissue_alpha
        Opacity of the tissue.
    cmap
        Color map to use.
    spot_size
        Size of the spot.
    show_color_bar
        Show color bar or not.
    show_axis
        Show axis or not.
    show_legend
        Show legend or not.
    dpi
        Set dpi as the resolution for the plot.
    name
        Name of the output figure file.
    output
        Save the figure as file or not.
    copy
        Return a copy instead of writing to adata.
    Returns
    -------
    Nothing
    """
    plt.rcParams['figure.dpi'] = dpi

    if str(cluster) not in adata.obs[use_label].unique():
        raise ValueError(
            "This cluster is non-exist, please choose another cluster!")

    if use_label not in adata.obs.columns:
        raise ValueError(
            "This label is non-exist, please choose another label!")
    plt.rcParams['figure.dpi'] = dpi

    colors = adata.obs[adata.obs[use_label]
                       == str(cluster)]["sub_cluster_labels"]
    keys = list(np.sort(colors.unique()))
    vals = np.arange(len(keys))
    mapping = dict(zip(keys, vals))

    # Option for turning off showing figure
    # plt.ioff()
    colors = colors.replace(mapping)

    # Initialize matplotlib
    fig, a = plt.subplots()

    # Plot scatter plot based on pixel of spots
    plot = a.scatter(adata.obs[adata.obs[use_label] == str(cluster)]["imagecol"],
                     adata.obs[adata.obs[use_label]
                               == str(cluster)]["imagerow"],
                     edgecolor="none", s=spot_size, marker="o", cmap=plt.get_cmap(cmap),
                     c=colors, alpha=data_alpha)

    if len(adata.obs[adata.obs[use_label] == str(cluster)]["sub_cluster_labels"].unique()) < 2:
        centroids = [centroidpython(adata.obs[adata.obs[use_label] == str(cluster)][[
                                    "imagecol", "imagerow"]].values)]
        classes = np.array(
            [adata.obs[adata.obs[use_label] == str(cluster)]["sub_cluster_labels"][0]])

    else:
        from sklearn.neighbors import NearestCentroid
        clf = NearestCentroid()
        clf.fit(adata.obs[adata.obs[use_label] == str(cluster)][["imagecol", "imagerow"]].values,
                adata.obs[adata.obs[use_label] == str(cluster)]["sub_cluster_labels"])

        centroids = clf.centroids_
        classes = clf.classes_

    norm = matplotlib.colors.Normalize(vmin=min(vals), vmax=max(vals))

    m = matplotlib.cm.ScalarMappable(norm=norm, cmap=cmap)

    for i, label in enumerate(classes):
        a.text(centroids[i][0], centroids[i][1], label, color='white', fontsize=5, zorder=3,
               bbox=dict(facecolor=matplotlib.colors.to_hex(m.to_rgba(mapping[label])),
                         boxstyle='round', alpha=0.5))

    if library_id is None:
        library_id = list(adata.uns["spatial"].keys())[0]

    image = adata.uns["spatial"][library_id]["images"][adata.uns["spatial"]["use_quality"]]
    
    a.imshow(image, alpha=tissue_alpha, zorder=-1,)


    if not show_axis:
        a.axis('off')

    plt.show()

    if output is not None:
        if name is None:
            print("The file name is not defined!")
            name = use_label
        fig.savefig(output + "/" + name + ".png", dpi=dpi,
                    bbox_inches='tight', pad_inches=0)

    # plt.close(fig)

    # if "plots" not in adata.uns:
    #    adata.uns['plots'] = {}
    #adata.uns['plots'].update({"subcl_" + use_label + "_" + str(cluster): fig_np})

    #print("The plot stored in adata.uns['plots']['" + "subcl_" + use_label + "_" + str(cluster)+ "']")

    #import matplotlib.pyplot as plt
    #plt.rcParams['figure.dpi'] = dpi
    #plt.imshow(adata.uns['plots']["subcl_" + use_label + "_" + str(cluster)])
    # plt.axis('off')


def centroidpython(data):
    x, y = zip(*data)
    l = len(x)
    return sum(x) / l, sum(y) / l



from mpl_toolkits.mplot3d import Axes3D
from matplotlib.patches import FancyArrowPatch
from mpl_toolkits.mplot3d import proj3d
import matplotlib.pyplot as plt
import numpy as np
from PIL import Image
import pandas as pd
import matplotlib
import numpy as np
import networkx as nx
from stlearn._compat import Literal
from typing import Optional, Union
from anndata import AnnData
import warnings

def local_plot(
    adata: AnnData,
    use_label: str = "louvain",
    use_cluster: int = None,
    reverse: bool = False,
    cluster: int = 0,
    data_alpha: float = 1.0,
    arrow_alpha: float = 1.0,
    branch_alpha: float = 1.0,
    spot_size: Union[float,int] = 1,
    show_color_bar: bool = True,
    show_axis: bool = False,
    dpi: int = 180,
    copy: bool = False,
) -> Optional[AnnData]:
    
    """\
    Local spatial trajectory inference plot.

    Parameters
    ----------
    adata
        Annotated data matrix.
    use_label
        Use label result of clustering method.
    use_cluster
        Choose a specific clusters that will display in the plot.
    data_alpha
        Opacity of the spot.
    arrow_alpha
        Opacity of the arrow.
    branch_alpha
        Opacity of the branch edge.
    edge_alpha
        Opacity of edge in PAGA graph in the tissue.
    node_alpha
        Opacity of node in PAGA graph in the tissue.
    spot_size
        Size of the spot.
    show_color_bar
        Show color bar or not.
    show_axis
        Show axis or not.
    show_legend
        Show legend or not.
    dpi
        Set dpi as the resolution for the plot.
    copy
        Return a copy instead of writing to adata.
    Returns
    -------
    Nothing
    """

    tmp=adata.obs[adata.obs[use_label]==str(use_cluster)]
    ref_cluster = adata[list(tmp.index)]

    plt.rcParams['figure.dpi'] = dpi
    plt.rcParams['figure.figsize'] = 5, 5
    fig = plt.figure()
    ax = fig.gca(projection='3d')

    centroids_ = []
    classes_ = []
    order_dict = {}
    order = 0
    for i in ref_cluster.obs["sub_cluster_labels"].unique():
        classes_.append(i)
        centroids_.append(adata.uns["centroid_dict"][int(i)])
        order_dict[int(i)] = int(order)
        order += 1

    stdm = adata.uns["ST_distance_matrix"]
    non_abs_dpt = adata.uns["nonabs_dpt_distance_matrix"]

    for i in range(0,len(centroids_)):
        if i == len(centroids_)-1:
            break
        j=0
        while j <= (len(centroids_)-2-i):
            j=j+1
            
            
            m = stdm[order_dict[int(classes_[i])],order_dict[int(classes_[i+j])]]
            dpt_distance = non_abs_dpt[order_dict[int(classes_[i])],order_dict[int(classes_[i+j])]]
            y = calculate_y(np.abs(m))
            
            x = np.linspace(centroids_[i][0],centroids_[i+j][0], 1000)
            z = np.linspace(centroids_[i][1],centroids_[i+j][1], 1000)
            
            
            branch = ax.plot(x,y,z,zorder=10,c="#333333",linewidth=1,alpha=branch_alpha)
            if reverse:
                dpt_distance = -dpt_distance
            if dpt_distance <=0:
                xyz = ([x[500],x[520]],[y[500],y[520]],[z[500],z[520]])
            else:
                xyz = ([x[520],x[500]],[y[520],y[500]],[z[520],z[500]])
            
            arrow = Arrow3D(xyz[0],xyz[1],xyz[2],zorder=10,mutation_scale=5, 
                   lw=1, arrowstyle="simple", color="r",alpha=arrow_alpha,)
            ax.add_artist(arrow)
            
            ax.text(x[500], y[500]-0.15, z[500], np.round(np.abs(m),3), color='black',size=5,zorder=100)
        
            
    sc = ax.scatter(ref_cluster.obs["imagecol"],
               0,
               ref_cluster.obs["imagerow"],
               c=ref_cluster.obs["dpt_pseudotime"],
               s=spot_size,cmap = "viridis",zorder=0, alpha=data_alpha)

    _ = ax.scatter(adata.obs[adata.obs[use_label]!=cluster]["imagecol"],
               0,
               adata.obs[adata.obs[use_label]!=cluster]["imagerow"],
               c="grey",
               s=spot_size,zorder=0,alpha=0.1)

    if show_color_bar:
        cb = fig.colorbar(sc,cax = fig.add_axes([0.1, 0.3, 0.03, 0.5]))
        cb.outline.set_visible(False)
        
    ax.set_ylim([-1,0])
    ax.set_xlim([min(adata.obs["imagecol"])-10,max(adata.obs["imagecol"])+10])
    ax.set_zlim([min(adata.obs["imagerow"])-10,max(adata.obs["imagerow"])+10])

    if not show_axis:
    # make the grid lines transparent
        ax.xaxis._axinfo["grid"]['color'] =  (1,1,1,0)
        ax.yaxis._axinfo["grid"]['color'] =  (1,1,1,0)
        ax.zaxis._axinfo["grid"]['color'] =  (1,1,1,0)
        ax.get_xaxis().set_ticks([])
        ax.get_yaxis().set_ticks([])
        ax.get_zaxis().set_ticks([])

    ax.invert_zaxis()
    ax.patch.set_visible(False) 
    plt.show()
    plt.rcParams['figure.figsize'] = 6, 4

def calculate_y(m):
    import math
    mu = 0
    variance = 1
    sigma = math.sqrt(variance)
    y = np.linspace(mu - 0.5*sigma, mu + 0.5*sigma, 1000)
    y = np.cos(np.absolute(y))
    y = -(m)*(y - np.min(y))/np.ptp(y)
    
    return y



class Arrow3D(FancyArrowPatch):
    def __init__(self, xs, ys, zs, *args, **kwargs):
        FancyArrowPatch.__init__(self, (0,0), (0,0), *args, **kwargs)
        self._verts3d = xs, ys, zs

    def draw(self, renderer):
        xs3d, ys3d, zs3d = self._verts3d
        xs, ys, zs = proj3d.proj_transform(xs3d, ys3d, zs3d, renderer.M)
        self.set_positions((xs[0],ys[0]),(xs[1],ys[1]))
        FancyArrowPatch.draw(self, renderer)


"""Reading and Writing
"""
from pathlib import Path, PurePath
from typing import Optional, Union
from anndata import AnnData
import numpy as np
from PIL import Image
import pandas as pd
import stlearn
import scanpy

def Read10X(
    path: Union[str, Path],
    genome: Optional[str] = None,
    count_file: str = "filtered_feature_bc_matrix.h5",
    library_id: str = None,
    load_images: Optional[bool] = True,
    use_quality: str = "hires"
    ) -> AnnData:
    
    """\
    Read Visium data from 10X (wrap read_visium from scanpy)

    In addition to reading regular 10x output,
    this looks for the `spatial` folder and loads images,
    coordinates and scale factors.
    Based on the `Space Ranger output docs`_.

    .. _Space Ranger output docs: https://support.10xgenomics.com/spatial-gene-expression/software/pipelines/latest/output/overview

    Parameters
    ----------
    path
        Path to directory for visium datafiles.
    genome
        Filter expression to genes within this genome.
    count_file
        Which file in the passed directory to use as the count file. Typically would be one of:
        'filtered_feature_bc_matrix.h5' or 'raw_feature_bc_matrix.h5'.
    library_id
        Identifier for the visium library. Can be modified when concatenating multiple adata objects.
    Returns
    -------
    Annotated data matrix, where observations/cells are named by their
    barcode and variables/genes by gene name. Stores the following information:
    :attr:`~anndata.AnnData.X`
        The data matrix is stored
    :attr:`~anndata.AnnData.obs_names`
        Cell names
    :attr:`~anndata.AnnData.var_names`
        Gene names
    :attr:`~anndata.AnnData.var`\\ `['gene_ids']`
        Gene IDs
    :attr:`~anndata.AnnData.var`\\ `['feature_types']`
        Feature types
    :attr:`~anndata.AnnData.uns`\\ `['spatial']`
        Dict of spaceranger output files with 'library_id' as key
    :attr:`~anndata.AnnData.uns`\\ `['spatial'][library_id]['images']`
        Dict of images (`'hires'` and `'lowres'`)
    :attr:`~anndata.AnnData.uns`\\ `['spatial'][library_id]['scalefactors']`
        Scale factors for the spots
    :attr:`~anndata.AnnData.uns`\\ `['spatial'][library_id]['metadata']`
        Files metadata: 'chemistry_description', 'software_version'
    :attr:`~anndata.AnnData.obsm`\\ `['spatial']`
        Spatial spot coordinates, usable as `basis` by :func:`~scanpy.pl.embedding`.
    """
    
    from scanpy import read_visium
    adata = read_visium(path, genome=None,
     count_file='filtered_feature_bc_matrix.h5',
      library_id=None,
       load_images=True)
    adata.var_names_make_unique()

    
    if library_id is None:
        library_id = list(adata.uns["spatial"].keys())[0]
        
    scale = adata.uns["spatial"][library_id]["scalefactors"]["tissue_"+use_quality+"_scalef"]
    image_coor = adata.obsm["spatial"]*scale

    adata.obs["imagecol"] = image_coor[:,0]
    adata.obs["imagerow"] = image_coor[:,1]
    adata.uns["spatial"]["use_quality"] = use_quality

    return adata


def ReadOldST(
    count_matrix_file: Union[str, Path] = None,
    spatial_file: Union[str, Path] = None,
    image_file: Union[str, Path] = None,
    ) -> AnnData:

    """\
    Read Old Spatial Transcriptomics data

    Parameters
    ----------
    count_matrix_file
        Path to count matrix file.
    spatial_file
        Path to spatial location file.
    image_file
        Path to the tissue image file
    Returns
    -------
    AnnData
    """

    adata = stlearn.read.file_table(count_matrix_file)
    adata=stlearn.add.parsing(adata,
        coordinates_file = spatial_file)
    stlearn.add.image(adata, imgpath=image_file)

    return adata


def ReadSlideSeq(
    count_matrix_file: Union[str, Path],
    spatial_file: Union[str, Path],
    library_id: str = None,
    ) -> AnnData:

    """\
    Read Slide-seq data

    Parameters
    ----------
    count_matrix_file
        Path to count matrix file.
    spatial_file
        Path to spatial location file.
    library_id
        Spatial data ID for storing related image data
    Returns
    -------
    AnnData
    """

    count = pd.read_csv(count_matrix_file)
    meta = pd.read_csv(spatial_file)

    adata = AnnData(count.iloc[:,1:].set_index("gene").T)

    adata.var["ENSEMBL"] = count["ENSEMBL"].values

    adata.obs["index"] = meta["index"].values
    adata.obs["imagecol"] = meta["x"].values/scale_factor
    adata.obs["imagerow"] = meta["y"].values/scale_factor

    # Create image
    max_size = np.max([adata.obs["imagecol"].max(),adata.obs["imagerow"].max()])
    max_size = int(max_size + 0.1*max_size)

    image = Image.new('RGB', (max_size, max_size), (255, 255, 255))
    imgarr = np.array(image)

    if library_id is None:
        library_id = "Slide-seq"

    adata.uns["spatial"] = {}
    adata.uns["spatial"][library_id] = {}
    adata.uns["spatial"][library_id]["images"] = {}
    adata.uns["spatial"][library_id]["images"]["hires"] = imgarr
    adata.uns["spatial"]["use_quality"] = "hires"
    adata.uns["spatial"][library_id]["scalefactors"] = {}
    adata.uns["spatial"][library_id]["scalefactors"]["tissue_hires_scalef"] = 1

    adata.obsm["spatial"] = meta[["x","y"]].values

    return adata

def ReadMERFISH(
    count_matrix_file: Union[str, Path],
    spatial_file: Union[str, Path],
    library_id: str = None,
    ) -> AnnData:

    """\
    Read MERFISH data

    Parameters
    ----------
    count_matrix_file
        Path to count matrix file.
    spatial_file
        Path to spatial location file.
    library_id
        Spatial data ID for storing related image data
    Returns
    -------
    AnnData
    """

    coordinates = pd.read_excel(spatial_file, index_col=0)
    if coordinates.min().min() < 0:
        coordinates = coordinates + np.abs(coordinates.min().min())+100
    from scanpy import read_csv
    counts = read_csv(count_matrix_file).transpose()

    adata_merfish = counts[coordinates.index, :]
    adata_merfish.obsm["spatial"] = coordinates.to_numpy()
    adata_merfish.obs["imagecol"] = adata_merfish.obsm["spatial"][:,0]
    adata_merfish.obs["imagerow"] = adata_merfish.obsm["spatial"][:,1]

    # Create image
    max_size = np.max([adata_merfish.obs["imagecol"].max(),adata_merfish.obs["imagerow"].max()])
    max_size = int(max_size + 0.1*max_size)
    image = Image.new('RGB', (max_size, max_size), (255, 255, 255))
    imgarr = np.array(image)


    if library_id is None:
        library_id = "MERSEQ"

    adata_merfish.uns["spatial"] = {}
    adata_merfish.uns["spatial"][library_id] = {}
    adata_merfish.uns["spatial"][library_id]["images"] = {}
    adata_merfish.uns["spatial"][library_id]["images"]["hires"] = imgarr
    adata_merfish.uns["spatial"]["use_quality"] = "hires"
    adata_merfish.uns["spatial"][library_id]["scalefactors"] = {}
    adata_merfish.uns["spatial"][library_id]["scalefactors"]["tissue_hires_scalef"] = 0.5

    return adata_merfish


"""Top-level package for stLearn."""

__author__ = """Genomics and Machine Learning lab"""
__email__ = 'duy.pham@uq.edu.au'
__version__ = '0.1.9'


from . import read
from . import add
from . import pp
from . import em
from . import tl
from . import pl
from . import view
from . import spatial

# Wrapper

from .wrapper.read import ReadSlideSeq
from .wrapper.read import Read10X
from .wrapper.read import ReadOldST
from .wrapper.read import ReadMERFISH
from .wrapper.clustering import SMEclust



from tensorflow.keras import backend as K


def encode(tiles, model):
    features = model.predict(tiles)
    features = features.ravel()
    return features


class Model:
    __name__ = "CNN base model"

    def __init__(self, base, batch_size=1):
        self.base = base
        self.model, self.preprocess = self.load_model()
        self.batch_size = batch_size
        self.data_format = K.image_data_format()

    def load_model(self):
        if self.base == "resnet50":
            from tensorflow.keras.applications.resnet50 import ResNet50, preprocess_input
            cnn_base_model = ResNet50(
                include_top=False, weights='imagenet', pooling="avg")
        elif self.base == "vgg16":
            from tensorflow.keras.applications.vgg16 import VGG16, preprocess_input
            cnn_base_model = VGG16(
                include_top=False, weights='imagenet', pooling="avg")
        elif self.base == "inception_v3":
            from tensorflow.keras.applications.inception_v3 import InceptionV3, preprocess_input
            cnn_base_model = InceptionV3(
                include_top=False, weights='imagenet', pooling="avg")
        elif self.base == "xception":
            from tensorflow.keras.applications.xception import Xception, preprocess_input
            cnn_base_model = Xception(
                include_top=False, weights='imagenet', pooling="avg")
        else:
            raise ValueError('{} is not a valid model'.format(self.base))
        return cnn_base_model, preprocess_input

    def predict(self, x):
        if self.data_format == "channels_first":
            x = x.transpose(0, 3, 1, 2)
        x = self.preprocess(x.astype(K.floatx()))
        return self.model.predict(x, batch_size=self.batch_size)
